# This file is part of Xpra.
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import socket
from time import sleep, time, monotonic
from queue import Queue
from typing import Any
from collections.abc import Callable, Iterable

from xpra.net.net_util import get_network_caps
from xpra.net.compression import Compressed, compressed_wrapper, MIN_COMPRESS_SIZE
from xpra.net.protocol.constants import CONNECTION_LOST
from xpra.net.common import MAX_PACKET_SIZE, PacketType
from xpra.net.digest import get_salt, gendigest
from xpra.scripts.config import parse_number, str_to_bool
from xpra.common import FULL_INFO, ConnectionMessage
from xpra.os_util import get_hex_uuid, gi_import
from xpra.exit_codes import ExitValue
from xpra.util.objects import typedict
from xpra.util.str_fn import csv, Ellipsizer, strtobytes, nicestr
from xpra.util.env import envint, envbool
from xpra.util.version import XPRA_VERSION, vparts
from xpra.server.core import get_server_info, get_thread_info, proto_crypto_caps
from xpra.log import Logger

GLib = gi_import("GLib")

log = Logger("proxy")

PROXY_QUEUE_SIZE = envint("XPRA_PROXY_QUEUE_SIZE", 0)
PASSTHROUGH_AUTH = envbool("XPRA_PASSTHROUGH_AUTH", True)

PING_INTERVAL = max(1, envint("XPRA_PROXY_PING_INTERVAL", 5)) * 1000
PING_WARNING = max(5, envint("XPRA_PROXY_PING_WARNING", 5))
PING_TIMEOUT = max(5, envint("XPRA_PROXY_PING_TIMEOUT", 90))

CLIENT_REMOVE_CAPS = ("cipher", "challenge", "digest", "aliases", "compression", "lz4", "lz0", "zlib")
CLIENT_REMOVE_CAPS_CHALLENGE = ("cipher", "digest", "aliases", "compression", "lz4", "lz0", "zlib")


def number(k, v) -> int:
    return parse_number(int, k, v)


SESSION_OPTION_WHITELIST: dict[str, Callable] = {
    "compression_level": number,
    "lz4": str_to_bool,
    "rencodeplus": str_to_bool,
}


def sanitize_session_options(options: dict[str, str]) -> dict[str, Any]:
    d = {}
    for k, v in options.items():
        parser = SESSION_OPTION_WHITELIST.get(k)
        if parser:
            log("trying to add %s=%s using %s", k, v, parser)
            try:
                d[k] = parser(k, v)
            except Exception as e:
                log.warn("failed to parse value %s for %s using %s: %s", v, k, parser, e)
    return d


def filter_caps(caps: typedict, prefixes: Iterable[str], proto=None) -> dict:
    # removes caps that overrides / does not use:
    pcaps = {}
    removed = []
    for k in caps.keys():
        if any(e for e in prefixes if str(k).startswith(e)):
            removed.append(k)
        else:
            pcaps[k] = caps[k]
    log("filtered out %s matching %s", removed, prefixes)
    # replace the network caps with the proxy's own:
    pcaps |= get_network_caps() | proto_crypto_caps(proto)
    # then add the proxy info:
    si = get_server_info()
    if FULL_INFO > 0:
        si["hostname"] = socket.gethostname()
    pcaps["proxy"] = si
    return pcaps


def replace_packet_item(packet: PacketType,
                        index: int,
                        new_value: Compressed | str | dict | int) -> PacketType:
    # make the packet data mutable and replace the contents at `index`:
    assert index > 0
    lpacket = list(packet)
    lpacket[index] = new_value
    # noinspection PyTypeChecker
    return tuple(lpacket)


# noinspection PyMethodMayBeStatic
class ProxyInstance:

    def __init__(self, session_options: dict[str, str],
                 pings: int,
                 disp_desc: dict, cipher: str, cipher_mode: str, encryption_key: bytes, caps: typedict):
        self.session_options = session_options
        self.pings = pings
        self.disp_desc = disp_desc
        self.cipher = cipher
        self.cipher_mode = cipher_mode
        self.encryption_key = encryption_key
        self.caps = caps
        log("ProxyInstance%s", (
            session_options,
            disp_desc, cipher, cipher_mode, encryption_key,
            "%s: %s.." % (type(caps), Ellipsizer(caps))))
        self.uuid: str = get_hex_uuid()
        self.client_protocol = None
        self.client_has_more = False
        self.server_protocol = None
        self.server_has_more = False
        # ping handling:
        self.client_last_ping: float = 0
        self.server_last_ping: float = 0
        self.client_last_ping_echo: float = 0
        self.server_last_ping_echo: float = 0
        self.client_last_ping_latency: float = 0
        self.server_last_ping_latency: float = 0
        self.client_ping_timer = 0
        self.server_ping_timer = 0
        self.client_challenge_packet: tuple = ()
        self.exit = False
        self.server_packets: Queue[PacketType] = Queue(PROXY_QUEUE_SIZE)
        self.client_packets: Queue[PacketType] = Queue(PROXY_QUEUE_SIZE)

    def is_alive(self) -> bool:
        return not self.exit

    def log_start(self) -> None:
        assert self.client_protocol and self.server_protocol
        log.info("started %s", self)
        log.info(" for client %s", self.client_protocol._conn)
        log.info(" and server %s", self.server_protocol._conn)

    def run(self) -> ExitValue:
        # server connection tweaks:
        self.server_protocol.large_packets += ["input-devices", "draw", "window-icon",
                                               "keymap-changed", "server-settings"]
        if self.caps.boolget("file-transfer"):
            self.server_protocol.large_packets += ["send-file", "send-file-chunk"]
            self.client_protocol.large_packets += ["send-file", "send-file-chunk"]
        self.server_protocol.set_compression_level(int(self.session_options.get("compression_level", 0)))
        self.server_protocol.enable_default_encoder()

        self.start_network_threads()
        self.schedule_client_ping()

        self.send_hello()
        return 0

    def start_network_threads(self) -> None:
        raise NotImplementedError()

    ################################################################################

    def close_connections(self, skip_proto, *reasons) -> None:
        for proto in (self.client_protocol, self.server_protocol):
            if proto and proto != skip_proto:
                log("sending disconnect to %s", proto)
                rstr = [nicestr(reason) for reason in reasons]
                proto.send_disconnect([ConnectionMessage.SERVER_SHUTDOWN] + rstr)
        # wait for connections to close down cleanly before we exit
        cp = self.client_protocol
        sp = self.server_protocol
        for i in range(10):
            if cp.is_closed() and sp.is_closed():
                return
            if i == 0:
                log("waiting for network connections to close")
            elif i == 1:
                log.info("waiting for network connections to close")
            else:
                log("still waiting %i/10 - client.closed=%s, server.closed=%s",
                    i + 1, cp.is_closed(), sp.is_closed())
            sleep(0.1)
        if not cp.is_closed():
            log("proxy instance client connection has not been closed yet:")
            log(" %s", cp)
            cp.close()
        if not sp.is_closed():
            log("proxy instance server connection has not been closed yet:")
            log(" %s", sp)
            sp.close()

    def send_disconnect(self, proto, *reasons) -> None:
        log("send_disconnect(%s, %s)", proto, reasons)
        if proto.is_closed():
            return
        proto.send_disconnect(reasons)
        self.timeout_add(1000, self.force_disconnect, proto)

    def force_disconnect(self, proto) -> None:
        proto.close()

    def stop(self, skip_proto, *reasons) -> None:
        log("stop(%s, %s)", skip_proto, reasons)
        if not self.exit:
            log.info("stopping %s", self)
            for x in reasons:
                log.info(" %s", x)
            self.exit = True
        self.cancel_client_ping_timer()
        self.cancel_server_ping_timer()
        self.close_connections(skip_proto, *reasons)
        self.stopped()

    def stopped(self) -> None:
        pass

    ################################################################################

    def get_proxy_info(self, proto) -> dict[str, Any]:
        sinfo = {"threads": get_thread_info(proto)}
        sinfo.update(get_server_info())
        linfo = {}
        if self.client_last_ping_latency:
            linfo["client"] = round(self.client_last_ping_latency)
        if self.server_last_ping_latency:
            linfo["server"] = round(self.server_last_ping_latency)
        return {
            "proxy": {
                "version": vparts(XPRA_VERSION, FULL_INFO + 1),
                "": sinfo,
                "latency": linfo,
            },
        }

    def get_connection_info(self) -> dict[str, Any]:
        return {
            "client": self.client_protocol.get_info(),
            "server": self.server_protocol.get_info(),
        }

    def get_info(self) -> dict[str, Any]:
        return {"connection": self.get_connection_info()}

    ################################################################################

    def send_hello(self, challenge_response=b"", client_salt=b"") -> None:
        hello = self.filter_client_caps()
        if challenge_response:
            hello.update({
                "challenge_response": challenge_response,
                "challenge_client_salt": client_salt,
            })
        hello.setdefault("network", {})["pings"] = self.pings
        self.queue_server_packet(("hello", hello))

    def filter_client_caps(self, remove=CLIENT_REMOVE_CAPS) -> dict:
        fc = filter_caps(self.caps, remove, self.server_protocol)
        # the display string may override the username:
        username = self.disp_desc.get("username")
        if username:
            fc["username"] = username
        # update with options provided via config if any:
        fc.update(sanitize_session_options(self.session_options))
        return fc

    def filter_server_caps(self, caps: typedict) -> dict:
        self.server_protocol.enable_encoder_from_caps(caps)
        return filter_caps(caps, ("aliases",), self.client_protocol)

    ################################################################################

    def queue_client_packet(self, packet: PacketType) -> None:
        log("queueing client packet: %s (queue size=%s)", packet[0], self.client_packets.qsize())
        self.client_packets.put(packet)
        self.client_protocol.source_has_more()

    def get_client_packet(self) -> tuple[PacketType, bool, bool]:
        # server wants a packet
        p = self.client_packets.get()
        s = self.client_packets.qsize()
        log("sending to client: %s (queue size=%i)", p[0], s)
        return p, True, s > 0 or self.server_has_more

    def process_client_packet(self, proto, packet: PacketType) -> None:
        packet_type = str(packet[0])
        log("process_client_packet: %s", packet_type)
        if packet_type == CONNECTION_LOST:
            self.stop(proto, "client connection lost")
            return
        self.client_has_more = proto.receive_pending
        if packet_type == "hello":
            if not self.client_challenge_packet:
                log.warn("Warning: invalid hello packet from client")
                log.warn(" received after initial authentication (dropped)")
                return
            log("forwarding client hello")
            log(" for challenge packet %s", self.client_challenge_packet)
            # update caps with latest hello caps from client:
            self.caps = typedict(packet[1])
            # keep challenge data in the hello response:
            hello = self.filter_client_caps(CLIENT_REMOVE_CAPS_CHALLENGE)
            self.queue_server_packet(("hello", hello))
            return
        if packet_type == "ping_echo" and self.client_ping_timer and len(packet) >= 7 and packet[6] == self.uuid:
            # this is one of our ping packets:
            self.client_last_ping_echo = float(packet[1])
            self.client_last_ping_latency = 1000 * monotonic() - self.client_last_ping_echo
            log("ping-echo: client latency=%.1fms", self.client_last_ping_latency)
            return
        # the packet types below are forwarded:
        if packet_type == "disconnect":
            reasons = tuple(str(x) for x in packet[1:])
            log("got disconnect from client: %s", csv(reasons))
            if self.exit:
                self.client_protocol.close()
            else:
                self.stop(None, "disconnect from client", *reasons)
        elif packet_type == "send-file" and packet[6]:
            packet = self.compressed_marker(packet, 6, "file-data")
        elif packet_type == "send-file-chunk" and packet[3]:
            packet = self.compressed_marker(packet, 3, "file-chunk-data")
        self.queue_server_packet(packet)

    def compressed_marker(self, packet: PacketType, index: int, description: str) -> PacketType:
        return replace_packet_item(packet, index, Compressed(description, packet[index]))

    def queue_server_packet(self, packet: PacketType) -> None:
        log("queueing server packet: %s (queue size=%s)", packet[0], self.server_packets.qsize())
        self.server_packets.put(packet)
        self.server_protocol.source_has_more()

    def get_server_packet(self):
        # server wants a packet
        p = self.server_packets.get()
        s = self.server_packets.qsize()
        log("sending to server: %s (queue size=%i)", p[0], s)
        return p, True, s > 0 or self.client_has_more

    def _packet_recompress(self, packet: PacketType, index: int, name: str) -> PacketType:
        if len(packet) <= index:
            return packet
        data = packet[index]
        if len(data) < MIN_COMPRESS_SIZE:
            return packet
        # this is ugly and not generic!
        kw = {"lz4": self.caps.boolget("lz4")}
        return replace_packet_item(packet, index, compressed_wrapper(name, data, can_inline=False, **kw))

    def cancel_server_ping_timer(self) -> None:
        spt = self.server_ping_timer
        log("cancel_server_ping_timer() server_ping_timer=%s", spt)
        if spt:
            self.server_ping_timer = 0
            self.source_remove(spt)

    def cancel_client_ping_timer(self) -> None:
        cpt = self.client_ping_timer
        log("cancel_client_ping_timer() client_ping_timer=%s", cpt)
        if cpt:
            self.client_ping_timer = 0
            self.source_remove(cpt)

    def schedule_server_ping(self) -> None:
        log("schedule_server_ping()")
        self.cancel_server_ping_timer()
        self.server_last_ping_echo = monotonic()
        self.server_ping_timer = self.timeout_add(PING_INTERVAL, self.send_server_ping)

    def schedule_client_ping(self) -> None:
        log("schedule_client_ping()")
        self.cancel_client_ping_timer()
        self.client_last_ping_echo = monotonic()
        self.client_ping_timer = self.timeout_add(PING_INTERVAL, self.send_client_ping)

    def send_server_ping(self) -> bool:
        log("send_server_ping() server_last_ping=%s", self.server_last_ping)
        # if we've already sent one, check for the echo:
        if self.server_last_ping:
            delta = self.server_last_ping - self.server_last_ping_echo
            if delta > PING_WARNING:
                log.warn("Warning: late server ping, %i seconds", delta)
            if delta > PING_TIMEOUT:
                log.error("Error: server ping timeout, %i seconds", delta)
                self.stop(None, "proxy to server ping timeout")
                return False
        now = monotonic()
        self.server_last_ping = now
        packet: PacketType = ("ping", int(now * 1000), int(time() * 1000), self.uuid)
        self.queue_server_packet(packet)
        return True

    def send_client_ping(self) -> bool:
        log("send_client_ping() client_last_ping=%s", self.client_last_ping)
        # if we've already sent one, check for the echo:
        if self.client_last_ping:
            delta = self.client_last_ping - self.client_last_ping_echo
            if delta > PING_WARNING:
                log.warn("Warning: late client ping, %i seconds", delta)
            if delta > PING_TIMEOUT:
                log.error("Error: client ping timeout, %i seconds", delta)
                self.stop(None, "proxy to client ping timeout")
                return False
        now = monotonic()
        self.client_last_ping = now
        packet: PacketType = ("ping", int(now * 1000), int(time() * 1000), self.uuid)
        self.queue_client_packet(packet)
        return True

    def process_server_packet(self, proto, packet: PacketType) -> None:
        packet_type = str(packet[0])
        log("process_server_packet: %s", packet_type)
        if packet_type == CONNECTION_LOST:
            self.stop(proto, "server connection lost")
            return
        self.server_has_more = proto.receive_pending
        if packet_type == "disconnect":
            reason = str(packet[1])
            log("got disconnect from server: %s", reason)
            if self.exit:
                self.server_protocol.close()
            else:
                self.stop(None, "disconnect from server", reason)
        elif packet_type == "hello":
            c = typedict(packet[1])
            self.schedule_server_ping()
            maxw, maxh = c.intpair("max_desktop_size") or (4096, 4096)
            caps = self.filter_server_caps(c)
            # add new encryption caps:
            if self.cipher:
                # pylint: disable=import-outside-toplevel
                from xpra.net.crypto import crypto_backend_init, new_cipher_caps, DEFAULT_PADDING
                crypto_backend_init()
                enc_caps = self.caps.dictget("encryption")
                padding_options = typedict(enc_caps or {}).strtupleget("padding.options", [DEFAULT_PADDING])
                caps["encryption"] = new_cipher_caps(self.client_protocol,
                                                     self.cipher, self.cipher_mode, self.encryption_key,
                                                     padding_options,
                                                     enc_caps.boolget("always-pad", False),
                                                     enc_caps.boolget("stream", True))
            # may need to bump packet size:
            proto.max_packet_size = max(MAX_PACKET_SIZE, maxw * maxh * 4 * 4)
            packet = ("hello", caps)
        elif packet_type == "ping_echo" and self.server_ping_timer and len(packet) >= 7 and packet[6] == self.uuid:
            # this is one of our ping packets:
            self.server_last_ping_echo = float(packet[1])
            self.server_last_ping_latency = 1000 * monotonic() - self.server_last_ping_echo
            log("ping-echo: server latency=%.1fms", self.server_last_ping_latency)
            return
        elif packet_type == "info-response":
            # adds proxy info:
            # note: this is only seen by the client application
            # "xpra info" is a new connection, which talks to the proxy server...
            info = packet[1]
            info.update(self.get_proxy_info(proto))
        elif packet_type == "draw":
            pixel_data = packet[7]
            if pixel_data and len(pixel_data) > 1024:
                packet = self.compressed_marker(packet, 7, "pixel-data")
        elif packet_type == "sound-data":
            if packet[2]:
                # best if we use raw packets for the actual sound-data chunk:
                packet = self.compressed_marker(packet, 2, "sound-data")
        # we do want to reformat cursor packets...
        # as they will have been uncompressed by the network layer already:
        elif packet_type == "cursor":
            # packet = ["cursor", "png", x, y, width, height, xhot, yhot, serial, pixels, name]
            # or:
            # packet = ["cursor", ""]
            if len(packet) >= 8:
                packet = self._packet_recompress(packet, 9, "cursor")
        elif packet_type == "window-icon":
            if not isinstance(packet[5], str):
                packet = self._packet_recompress(packet, 5, "icon")
        elif packet_type == "send-file":
            if packet[6]:
                packet = self.compressed_marker(packet, 6, "file-data")
        elif packet_type == "send-file-chunk":
            if packet[3]:
                packet = self.compressed_marker(packet, 3, "file-chunk-data")
        elif packet_type == "challenge":
            password = self.disp_desc.get("password", self.session_options.get("password"))
            log("password from %s / %s = %s", self.disp_desc, self.session_options, password)
            if not password:
                if not PASSTHROUGH_AUTH:
                    self.stop(None, "authentication requested by the server,",
                              "but no password is available for this session")
                # otherwise, just forward it to the client
                self.client_challenge_packet = packet
            else:
                # client may have already responded to the challenge,
                # so we have to handle authentication from this end
                server_salt = strtobytes(packet[1])
                length = len(server_salt)
                digest = str(packet[3])
                salt_digest = "xor"
                if len(packet) >= 5:
                    salt_digest = str(packet[4])
                if salt_digest in ("xor", "des"):
                    self.stop(None, f"server uses legacy salt digest {salt_digest!r}")
                    return
                if salt_digest == "xor":
                    # with xor, we have to match the size
                    if length < 16:
                        raise ValueError(f"server salt is too short: only {length} bytes, minimum is 16")
                    if length > 256:
                        raise ValueError(f"server salt is too long: {length} bytes, maximum is 256")
                else:
                    # other digest, 32 random bytes is enough:
                    length = 32
                client_salt = get_salt(length)
                salt = gendigest(salt_digest, client_salt, server_salt)
                challenge_response = gendigest(digest, password, salt)
                if not challenge_response:
                    log("invalid digest module '%s': %s", digest)
                    self.stop(None, f"server requested {digest!r} digest but it is not supported")
                    return
                log.info("sending %s challenge response", digest)
                self.send_hello(challenge_response, client_salt)
                return
        self.queue_client_packet(packet)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2024 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from PyQt6.QtCore import Qt, QPoint, QEvent
from PyQt6.QtGui import QImage, QPixmap, QPainter, QKeyEvent
from PyQt6.QtWidgets import QMainWindow, QLabel, QSizePolicy

from xpra.client.qt6.keys import key_names
from xpra.log import Logger

log = Logger()


def get_pointer_position(event) -> tuple[int, int, int, int]:
    glo_pos = event.globalPosition()
    pos = event.position()
    return round(glo_pos.x()), round(glo_pos.y()), round(pos.x()), round(pos.y())


def get_event_pos(event) -> tuple[int, int]:
    pos = event.pos()
    return round(pos.x()), round(pos.y())


def get_modifiers(event: QKeyEvent) -> list[str]:
    modifiers = []
    mods = event.modifiers()
    for value, name in {
        Qt.KeyboardModifier.ShiftModifier: "shift",
        Qt.KeyboardModifier.ControlModifier: "control",
        Qt.KeyboardModifier.AltModifier: "alt",
        Qt.KeyboardModifier.MetaModifier: "meta",
        # Qt.KeyboardModifier.KeypadModifier: "numlock",
        # Qt.KeyboardModifier.KeypadModifier: "group",
    }.items():
        if mods & value:
            modifiers.append(name)
    return modifiers


class DrawingArea(QLabel):

    def __init__(self, client, wid):
        super().__init__()
        self.client = client
        self.send = client.send
        self.wid = wid
        self.seq = 0

    def mouseMoveEvent(self, event) -> None:
        pos = get_pointer_position(event)
        self.send("pointer", -1, self.seq, self.wid, pos, {})
        self.seq += 1

    def send_mouse_button_event(self, event, pressed=True) -> None:
        props = {
            # "modifiers": (),
            # "buttons": (),
        }
        button = {
            Qt.MouseButton.LeftButton: 1,
            Qt.MouseButton.RightButton: 2,
            Qt.MouseButton.MiddleButton: 3,
        }.get(event.button(), -1)
        if button < 0:
            return
        pos = get_pointer_position(event)
        self.send("pointer-button", -1, self.seq, self.wid, button, pressed, pos, props)
        self.seq += 1

    def mousePressEvent(self, event) -> None:
        self.send_mouse_button_event(event, True)

    def mouseReleaseEvent(self, event) -> None:
        self.send_mouse_button_event(event, False)

    def mouseDoubleClickEvent(self, event) -> None:
        self.send_mouse_button_event(event, True)
        self.send_mouse_button_event(event, True)

    def wheelEvent(self, event) -> None:
        pos = get_pointer_position(event)
        delta = event.pixelDelta()
        props = {}
        if delta.y() != 0:
            button = 4 if delta.y() > 0 else 5
            self.send("pointer-button", -1, self.seq, self.wid, button, True, pos, props)
            self.send("pointer-button", -1, self.seq, self.wid, button, False, pos, props)
            self.seq += 1
        if delta.x() != 0:
            button = 6 if delta.y() > 0 else 7
            self.send("pointer-button", -1, self.seq, self.wid, button, True, pos, props)
            self.send("pointer-button", -1, self.seq, self.wid, button, False, pos, props)
            self.seq += 1


class ClientWindow(QMainWindow):
    def __init__(self, client, wid: int, x: int, y: int, w: int, h: int, metadata: dict):
        super().__init__()
        self.client = client
        self.send = client.send
        self.wid = wid
        self.metadata = metadata
        if metadata.get("override-redirect"):
            self.setWindowFlags(Qt.WindowType.FramelessWindowHint | Qt.WindowType.Window | Qt.WindowType.X11BypassWindowManagerHint)
        self.setWindowTitle(metadata.get("title", ""))
        self.label = DrawingArea(client, wid)
        self.label.setSizePolicy(QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding))
        self.label.setMinimumSize(1, 1)
        self.canvas = QPixmap(w, h)
        self.canvas.fill(Qt.GlobalColor.white)
        self.label.setPixmap(self.canvas)
        self.setCentralWidget(self.label)
        self.move(x, y)
        self.resize(w, h)
        self.label.setMouseTracking(True)
        # self.setFixedSize(w, h)
        self.installEventFilter(self)

    def get_canvas_size(self) -> tuple[int, int]:
        size = self.label.size()
        return size.width(), size.height()

    def eventFilter(self, obj, event) -> bool:
        etype = event.type()
        if etype == QEvent.Type.WindowDeactivate:
            self.client.update_focus(0)
        elif etype == QEvent.Type.WindowActivate:
            self.client.update_focus(self.wid)
        elif etype == QEvent.Type.Show:
            x = self.pos().x() + self.label.pos().x()
            y = self.pos().y() + self.label.pos().y()
            w, h = self.get_canvas_size()
            if not self.metadata.get("override-redirect"):
                self.send("map-window", self.wid, x, y, w, h)
        elif etype == QEvent.Type.Hide:
            log("hidden - iconified?")
        elif etype == QEvent.Type.Move:
            x, y = get_event_pos(event)
            x += self.label.pos().x()
            y += self.label.pos().y()
            w, h = self.get_canvas_size()
            state = {}
            props = {}
            self.send("configure-window", self.wid, x, y, w, h, props, 0, state, False)
        elif etype == QEvent.Type.WindowStateChange:
            new_state = self.windowState()
            old_state = event.oldState()
            changes = {}
            for mask, name in {
                Qt.WindowState.WindowMinimized: "iconified",
                Qt.WindowState.WindowMaximized: "maximized",
                Qt.WindowState.WindowFullScreen: "fullscreen",
                Qt.WindowState.WindowActive: "focused",
            }.items():
                if new_state & mask != old_state & mask:
                    changes[name] = bool(new_state & mask)
            log(f"state changes: {changes}")
            if changes:
                props = {}
                self.send("configure-window", self.wid, 0, 0, 0, 0, props, 0, changes, True)
        elif etype == QEvent.Type.Resize:
            size = event.size()
            x = self.pos().x() + self.label.pos().x()
            y = self.pos().y() + self.label.pos().y()
            w = size.width()
            h = size.height()
            state = {}
            props = {}
            self.send("configure-window", self.wid, x, y, w, h, props, 0, state, False)
            old_pixmap = self.canvas
            self.canvas = QPixmap(w, h)
            self.canvas.fill(Qt.GlobalColor.white)
            painter = QPainter(self.canvas)
            painter.drawPixmap(0, 0, old_pixmap)
            painter.end()
            self.label.setPixmap(self.canvas)
        log(f"{obj}: {event} {event.type().name}")
        return False

    def keyPressEvent(self, event: QKeyEvent) -> None:
        self.send_key_event(event, True)

    def keyReleaseEvent(self, event: QKeyEvent) -> None:
        self.send_key_event(event, False)

    def closeEvent(self, event) -> None:
        log(f"close: {event}")
        self.send("close-window", self.wid)
        event.ignore()

    def send_key_event(self, event, pressed=True) -> None:
        keyval = event.key()
        name = key_names.get(keyval, event.text())
        keycode = event.nativeScanCode()
        string = event.text()
        group = 0
        modifiers = get_modifiers(event)
        native_modifiers = event.nativeModifiers()
        vk = event.nativeVirtualKey()
        log(f"key: {name!r}, {keyval=}, {keycode=}, {modifiers=} {native_modifiers=}, {vk=}")
        self.send("key-action", self.wid, name, pressed, modifiers, keyval, string, keycode, group)

    def draw(self, x, y, w, h, coding, data, stride) -> None:
        if coding in ("png", "jpg", "webp"):
            from PIL import Image
            from io import BytesIO
            img = Image.open(BytesIO(data))
            if img.mode == "RGBA":
                coding = "rgb32"
                stride = w * 4
            else:
                assert img.mode == "RGB"
                coding = "rgb24"
                stride = w * 3
            data = img.tobytes("raw", img.mode)
        assert coding in ("rgb24", "rgb32")
        fmt = QImage.Format.Format_RGB888 if coding == "rgb24" else QImage.Format.Format_ARGB32_Premultiplied
        image = QImage(data, w, h, stride, fmt)
        canvas = self.label.pixmap()
        painter = QPainter(canvas)
        point = QPoint(x, y)
        painter.drawImage(point, image)
        painter.end()
        self.label.setPixmap(canvas)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import sys
import time
import unittest

from xpra.util.objects import AdHocStruct
from xpra.os_util import WIN32, OSX
from xpra.util.env import OSEnvContext

from unit.test_util import silence_info
from unit.server.mixins.servermixintest_util import ServerMixinTest


class NotificationForwarderMixinTest(ServerMixinTest):

    def test_notification(self):
        with OSEnvContext():
            #remove any existing dbus environment so we don't pollute it:
            for k in tuple(os.environ.keys()):
                if k.startswith("DBUS"):
                    del os.environ[k]
            #start a dbus server:
            from xpra.server.dbus.start import start_dbus
            dbus_pid, dbus_env = start_dbus("dbus-launch --sh-syntax --close-stderr")
            try:
                if dbus_env:
                    os.environ.update(dbus_env)

                from xpra.server.mixins import notification
                opts = AdHocStruct()
                opts.notifications = "yes"
                with silence_info(notification):
                    self._test_mixin_class(notification.NotificationForwarder, opts)
                self.verify_packet_error(("notification-close", 1, "test", "hello"))
                self.verify_packet_error(("notification-action", 1))
                self.handle_packet(("notification-status", False))
                self.mixin.cleanup()
                time.sleep(0.1)
            finally:
                if dbus_pid:
                    import signal
                    os.kill(dbus_pid, signal.SIGINT)

def main():
    if WIN32 or OSX:
        print("skipping test on %s" % sys.platform)
    else:
        unittest.main()


if __name__ == '__main__':
    main()



#!/bin/env python3
# -*- coding: utf-8 -*-
# This file is part of Xpra.
# Copyright (C) 2024 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import re
import sys
import json
import os.path
import shlex
import hashlib
from typing import Any

from sysconfig import get_path
from datetime import datetime
from collections import namedtuple
from collections.abc import Iterable
from importlib.util import find_spec, spec_from_file_location, module_from_spec

from glob import glob
from subprocess import getstatusoutput, check_output, Popen, PIPE
from shutil import which, rmtree, copyfile, move, copytree

KEY_FILE = "E:\\xpra.pfx"
DIST = "dist"
LIB_DIR = f"{DIST}/lib"

ARCH = os.environ.get("MSYSTEM_CARCH", "")

DEBUG = os.environ.get("XPRA_DEBUG", "0") != "0"
PYTHON = os.environ.get("PYTHON", "python%i.%i" % sys.version_info[:2])
MINGW_PREFIX = os.environ.get("MINGW_PREFIX", "")
MSYSTEM_CARCH = os.environ.get("MSYSTEM_CARCH", "x86_64")
PACKAGE_PREFIX = os.environ.get("MINGW_PACKAGE_PREFIX", f"mingw-w64-{MSYSTEM_CARCH}") + "-"
MSYS2_PACKAGE_PREFIX = "msys2-"
MSYS_DLL_PREFIX = "msys-"


TIMESTAMP_SERVER = "http://timestamp.digicert.com"
# alternative:
# http://timestamp.comodoca.com/authenticode

PROGRAMFILES = "C:\\Program Files"
PROGRAMFILES_X86 = "C:\\Program Files (x86)"
SYSTEM32 = "C:\\Windows\\System32"

SRC_INFO = "xpra/src_info.py"
BUILD_INFO = "xpra/build_info.py"
SBOM_JSON = "sbom.json"

LOG_DIR = "packaging/MSWindows/"

NPROCS = int(os.environ.get("NPROCS", os.cpu_count()))

BUILD_CUDA_KERNEL = "packaging\\MSWindows\\BUILD_CUDA_KERNEL.BAT"

EXTRA_PYTHON_MODULES = [
    "unittest", "psutil", "pynvml",
    "browser_cookie3",
    "gssapi", "ldap", "ldap3", "pyu2f", "sqlite3",
]


def parse_command_line(argv: list[str]):
    from argparse import ArgumentParser, BooleanOptionalAction
    ap = ArgumentParser()

    # noinspection PyShadowingBuiltins
    def add(name: str, help: str, default=True):
        ap.add_argument(f"--{name}", default=default, action=BooleanOptionalAction, help=help)
    add("verbose", help="print extra diagnostic messages", default=False)
    add("clean", help="clean build directories")
    add("build", help="compile the source")
    add("install", help="run install step")
    add("fixups", help="run misc fixups")
    add("zip", help="generate a ZIP installation file")
    add("verpatch", help="run `verpatch` on the executables")
    add("light", help="trimmed down build")
    add("sbom", help="record SBOM")
    add("exe", help="create an EXE installer")
    add("run", help="run the EXE installer")
    add("msi", help="create an MSI installer")
    add("sign", help="sign the EXE and MSI installers")
    add("tests", help="run the unit tests", default=False)
    add("zip-modules", help="zip up python modules")
    add("cuda", help="build CUDA kernels for nvidia codecs")
    add("service", help="build the system service", default=ARCH != "aarch64")
    add("docs", help="generate the documentation", default=ARCH != "aarch64")
    add("html5", help="bundle the `xpra-html5` client")
    add("manual", help="bundle the user manual")
    add("numpy", help="bundle `numpy`")
    add("putty", help="bundle putty `plink`")
    add("openssh", help="bundle the openssh client")
    add("openssl", help="bundle the openssl tools")
    add("paexec", help="bundle `paexec`")
    add("desktop-logon", help="build `desktop-logon` tool")
    ap.add_argument("--build-args", default="", help="extra build arguments")

    args, unknown_args = ap.parse_known_args(argv)
    if args.light:
        # disable many switches:
        args.cuda = args.numpy = args.service = args.docs = args.html5 = args.manual = False
        args.putty = args.openssh = args.openssl = args.paexec = args.desktop_logon = False
    if args.verbose:
        global DEBUG
        DEBUG = True
    return args, unknown_args


def step(message: str) -> None:
    now = datetime.now()
    ts = f"{now.hour:02}:{now.minute:02}:{now.second:02}"
    print(f"* {ts} {message}")


def debug(message: str) -> None:
    if DEBUG:
        print("    "+message)


def csv(values: Iterable) -> str:
    return ", ".join(str(x) for x in values)


def du(path: str) -> int:
    if os.path.isfile(path):
        return os.path.getsize(path)
    return sum(os.path.getsize(f) for f in glob(f"{path}/**", recursive=True) if os.path.isfile(f))


def get_build_args(args) -> list[str]:
    xpra_args = ["--without-gstreamer_video"]
    if args.light:
        for option in (
            "shadow", "server", "proxy", "rfb",
            "dbus",
            "encoders", "avif",
            "nvfbc", "cuda_kernels",
            "csc_cython",
            "webcam",
            "win32_tools",
            "docs",
            "qt6_client",
            "websockets_browser_cookie",
            "service",
            "yaml",
        ):
            xpra_args.append(f"--without-{option}")
        xpra_args.append("--with-Os")
    else:
        xpra_args.append("--with-qt6_client")
    if not args.cuda:
        xpra_args.append("--without-nvidia")
    if args.build_args:
        xpra_args += shlex.split(args.build_args)
    # we can't do 'docs' this way :(
    # for arg in ("docs", ):
    #    value = getattr(args, arg)
    #    xpra_args.append(f"--with-{arg}={value}")       #ie: "--with-docs=True"
    return xpra_args


def _find_command(name: str, env_name: str, *paths) -> str:
    cmd = os.environ.get(env_name, "")
    if cmd and os.path.exists(cmd):
        return cmd
    cwd_cmd = os.path.abspath(f"./{name}.exe")
    if os.path.exists(cwd_cmd):
        return cwd_cmd
    cmd = which(name)
    if cmd and os.path.exists(cmd):
        return cmd
    for path in paths:
        if os.path.exists(path) and os.path.isfile(path):
            return path
    return ""


def find_command(name: str, env_name: str, *paths) -> str:
    cmd = _find_command(name, env_name, *paths)
    if cmd:
        return cmd
    print(f"{name!r} not found")
    print(f" (you can set the {env_name!r} environment variable to point to it)")
    print(f" tried %PATH%={os.environ.get('PATH')}")
    print(f" tried {paths=}")
    raise RuntimeError(f"{name!r} not found")


def search_command(wholename: str, *dirs: str) -> str:
    debug(f"searching for {wholename!r} in {dirs}")
    for dirname in dirs:
        if not os.path.exists(dirname):
            continue
        cmd = ["find", dirname, "-wholename", wholename]
        r, output = getstatusoutput(cmd)
        debug(f"getstatusoutput({cmd})={r}, {output}")
        if r == 0:
            return output.splitlines()[0]
    raise RuntimeError(f"{wholename!r} not found in {dirs}")


def find_java() -> str:
    try:
        return _find_command("java", "JAVA")
    except RuntimeError as e:
        debug(f"`java` was not found: {e}")
    # try my hard-coded default first to save time:
    java = f"{PROGRAMFILES}\\Java\\jdk1.8.0_121\\bin\\java.exe"
    if java and getstatusoutput(f"{java} --version")[0] == 0:
        return java
    dirs = (f"{PROGRAMFILES}/Java", f"{PROGRAMFILES}", f"{PROGRAMFILES_X86}")
    for directory in dirs:
        r, output = getstatusoutput(f"find {directory!r} -name java.exe")
        if r == 0:
            return output[0]
    raise RuntimeError(f"java.exe was not found in {dirs}")


def check_html5() -> None:
    step("Verify `xpra-html5` is installed")
    if not os.path.exists("xpra-html5") or not os.path.isdir("xpra-html5"):
        print("html5 client not found")
        print(" perhaps run: `git clone https://github.com/Xpra-org/xpra-html5`")
        raise RuntimeError("`xpra-html5` client not found")

    # Find a java interpreter we can use for the html5 minifier
    os.environ["JAVA"] = find_java()


def check_signtool() -> None:
    step("locating `signtool`")
    if os.path.exists("./signtool.exe"):
        return
    try:
        signtool = find_command("signtool", "SIGNTOOL",
                                f"{PROGRAMFILES}\\Microsoft SDKs\\Windows\\v7.1\\Bin\\signtool.exe"
                                f"{PROGRAMFILES}\\Microsoft SDKs\\Windows\\v7.1A\\Bin\\signtool.exe"
                                f"{PROGRAMFILES_X86}\\Windows Kits\\8.1\\Bin\\x64\\signtool.exe"
                                f"{PROGRAMFILES_X86}\\Windows Kits\\10\\App Certification Kit\\signtool.exe")
    except RuntimeError:
        signtool = ""
    if not signtool:
        # try the hard (slow) way:
        signtool = find_vs_command("signtool.exe")
        if not signtool:
            raise RuntimeError("signtool not found")
    debug(f"{signtool=}")
    copyfile(signtool, "./signtool.exe")


def show_tail(filename: str) -> None:
    if os.path.exists(filename):
        print(f"showing the last 10 lines of {filename!r}:")
        os.system(f"tail -n 10 {filename}")


def command_args(cmd: str | list[str]) -> list[str]:
    # make sure we use an absolute path for the command:
    if isinstance(cmd, str):
        parts = shlex.split(cmd)
    else:
        parts = cmd
    cmd_exe = parts[0]
    if not os.path.isabs(cmd_exe):
        cmd_exe = which(cmd_exe)
        if cmd_exe:
            parts[0] = cmd_exe
    return parts


def log_command(cmd: str | list[str], log_filename: str, **kwargs) -> None:
    debug(f"running {cmd!r} and sending the output to {log_filename!r}")
    if not os.path.isabs(log_filename):
        log_filename = os.path.join(LOG_DIR, log_filename)
    delfile(log_filename)
    if not kwargs.get("shell"):
        cmd = command_args(cmd)
    with open(log_filename, "w") as f:
        ret = Popen(cmd, stdout=f, stderr=f, **kwargs).wait()
    if ret != 0:
        show_tail(log_filename)
        raise RuntimeError(f"{cmd!r} failed and returned {ret}, see {log_filename!r}")


def find_delete(path: str, name: str, mindepth=0) -> None:
    debug(f"deleting all instances of {name!r} from {path!r}")
    cmd = ["find", path]
    if mindepth > 0:
        cmd += ["-mindepth", str(mindepth)]
    if name:
        cmd += ["-name", "'"+name+"'"]
    cmd += ["-type", "f"]
    cmd = command_args(cmd)
    output = check_output(cmd).decode()
    for filename in output.splitlines():
        delfile(filename)


def rmrf(path: str) -> None:
    if not os.path.exists(path):
        print(f"Warning: {path!r} does not exist")
        return
    rmtree(path)


def delfile(path: str) -> None:
    if os.path.exists(path):
        debug(f"removing {path!r}")
        os.unlink(path)


def clean() -> None:
    step("Cleaning output directories and generated files")
    debug("cleaning log files:")
    find_delete("packaging/MSWindows/", "*.log")
    for dirname in (DIST, "build"):
        rmrf(dirname)
        os.mkdir(dirname)
    # clean sometimes errors on removing pyd files,
    # so do it with rm instead:
    debug("removing compiled dll and pyd files:")
    find_delete("xpra", "*-cpython-*dll")
    find_delete("xpra", "*-cpython-*pyd")
    debug("python clean")
    log_command(f"{PYTHON} ./setup.py clean", "clean.log")
    debug("removing comtypes cache")
    # clean comtypes cache - it should not be included!
    check_output(command_args("clear_comtypes_cache.exe -y"))
    debug("ensure build info is regenerated")
    delfile(BUILD_INFO)


def find_windowskit_command(name="mc") -> str:
    debug(f"find_windowskit_command({name!r})")
    cwd_cmd = os.path.abspath(f"./{name}.exe")
    if os.path.exists(cwd_cmd):
        return cwd_cmd
    # the proper way would be to run vsvars64.bat
    # but we only want to locate 3 commands,
    # so we find them "by hand":
    ARCH_DIRS = ("x64", "x86")
    paths = []
    for prog_dir in (PROGRAMFILES, PROGRAMFILES_X86):
        for V in (8.1, 10):
            for ARCH in ARCH_DIRS:
                paths += glob(f"{prog_dir}\\Windows Kits\\{V}\\bin\\*\\{ARCH}\\{name}.exe")
    env_name = name.upper()   # ie: "MC"
    return find_command(name, env_name, *paths)


def find_vs_command(name="link") -> str:
    debug(f"find_vs_command({name!r})")
    cwd_cmd = os.path.abspath(f"./{name}.exe")
    if os.path.exists(cwd_cmd):
        return cwd_cmd
    dirs = []
    for prog_dir in (PROGRAMFILES, PROGRAMFILES_X86):
        for VSV in (14.0, 17.0, 19.0, 2019, 2022):
            vsdir = f"{prog_dir}\\Microsoft Visual Studio\\{VSV}"
            if os.path.exists(vsdir):
                dirs.append(f"{vsdir}\\VC\\bin")
                dirs.append(f"{vsdir}\\BuildTools\\VC\\Tools\\MSVC")
    return search_command(f"*/x64/{name}.exe", *dirs)


def build_service() -> None:
    step("Compiling system service shim")
    XPRA_SERVICE_EXE = "Xpra-Service.exe"
    SERVICE_SRC_DIR = os.path.join(os.path.abspath("."), "packaging", "MSWindows", "service")
    for filename in ("event_log.rc", "event_log.res", "MSG00409.bin", XPRA_SERVICE_EXE):
        path = os.path.join(SERVICE_SRC_DIR, filename)
        delfile(path)

    MC = find_windowskit_command("mc")
    RC = find_windowskit_command("rc")
    LINK = find_vs_command("link")

    log_command([MC, "-U", "event_log.mc"], "service-mc.log", cwd=SERVICE_SRC_DIR)
    log_command([RC, "event_log.rc"], "service-rc.log", cwd=SERVICE_SRC_DIR)
    log_command([LINK, "-dll", "-noentry", "-out:event_log.dll", "event_log.res"], "service-link.log",
                cwd=SERVICE_SRC_DIR)
    log_command(["g++", "-o", XPRA_SERVICE_EXE, "Xpra-Service.cpp", "-Wno-write-strings"], "service-gcc.log",
                cwd=SERVICE_SRC_DIR)
    os.rename(os.path.join(SERVICE_SRC_DIR, XPRA_SERVICE_EXE), os.path.join(DIST, XPRA_SERVICE_EXE))


VersionInfo = namedtuple("VersionInfo", ("string", "value", "revision", "full_string", "arch_info", "extra", "padded"))
version_info = VersionInfo("invalid", (0, 0), 0, "invalid", "arch", "extra", (0, 0, 0, 0))


def set_version_info(light: bool) -> None:
    step("Collecting version information")
    for filename in ("src_info.py", "build_info.py"):
        path = os.path.join("xpra", filename)
        delfile(path)
    log_command([PYTHON, "fs/bin/add_build_info.py", "src", "build"], "add-build-info.log")
    print("    Python " + sys.version)
    load_version_info(light)


def load_version_info(light: bool) -> None:

    def load_module(src: str):
        spec = spec_from_file_location("xpra", src)
        module = module_from_spec(spec)
        spec.loader.exec_module(module)
        return module

    xpra = load_module("xpra/__init__.py")
    src_info = load_module(SRC_INFO)

    revision = src_info.REVISION

    full_string = f"{xpra.__version__}-r{revision}"
    if src_info.LOCAL_MODIFICATIONS:
        full_string += "M"

    extra = "-Light" if light else ""
    # ie: "x86_64"
    arch_info = "-" + MSYSTEM_CARCH

    # for msi and verpatch:
    padded = (list(xpra.__version_info__) + [0, 0, 0])[:3] + [revision]
    padded = ".".join(str(x) for x in padded)

    print(f"    Xpra{extra} {full_string}")
    print(f"    using {NPROCS} cpus")
    global version_info
    version_info = VersionInfo(xpra.__version__, xpra.__version_info__, revision, full_string, arch_info, extra, padded)


################################################################################
# Build: clean, build extensions, generate exe directory


def build_cuda_kernels() -> None:
    step("Building CUDA kernels")
    for cupath in glob("fs/share/xpra/cuda/*.cu"):
        kname = os.path.splitext(os.path.basename(cupath))[0]
        cu = os.path.splitext(cupath)[0]
        # ie: "fs/share/xpra/cuda/BGRX_to_NV12.cu" -> "fs/share/xpra/cuda/BGRX_to_NV12"
        fatbin = f"{cu}.fatbin"
        if not os.path.exists(fatbin):
            debug(f"rebuilding {kname!r}: {fatbin!r} does not exist")
        else:
            ftime = os.path.getctime(fatbin)
            ctime = os.path.getctime(cupath)
            if ftime >= ctime:
                debug(f"{fatbin!r} ({ftime}) is already newer than {cupath!r} ({ctime})")
                continue
            debug(f"need to rebuild: {fatbin!r} ({ftime}) is older than {cupath!r} ({ctime})")
            os.unlink(fatbin)
        log_command([BUILD_CUDA_KERNEL, kname], f"nvcc-{kname}.log", shell=True)


def build_ext(args) -> None:
    step("Building Cython modules")
    build_args = get_build_args(args) + ["--inplace"]
    if NPROCS > 0:
        build_args += ["-j", str(NPROCS)]
    args_str = " ".join(build_args)
    log_command(f"{PYTHON} ./setup.py build_ext {args_str}", "build.log")


def run_tests() -> None:
    step("Running unit tests")
    env = os.environ.copy()
    env["PYTHONPATH"] = ".:./tests/unittests"
    env["XPRA_COMMAND"] = "./fs/bin/xpra"
    log_command(f"{PYTHON} ./setup.py unittests", "unittest.log", env=env)


def install_exe(args) -> None:
    step("Generating installation directory")
    args_str = " ".join(get_build_args(args))
    log_command(f"{PYTHON} ./setup.py install_exe {args_str} --install={DIST}", "install.log")


def install_docs() -> None:
    step("Generating the documentation")
    if not os.path.exists(f"{DIST}/doc"):
        os.mkdir(f"{DIST}/doc")
    env = os.environ.copy()
    env["PANDOC"] = find_command("pandoc", "PANDOC", f"{PROGRAMFILES}\\Pandoc\\pandoc.exe")
    log_command(f"{PYTHON} ./setup.py doc", "pandoc.log", env=env)


def fixups(light: bool) -> None:
    step("Fixups: paths, etc")
    # fix case sensitive mess:
    gi_dir = f"{LIB_DIR}/girepository-1.0"
    debug("Glib misspelt")
    os.rename(f"{gi_dir}/Glib-2.0.typelib", f"{gi_dir}/GLib-2.0.typelib.tmp")
    os.rename(f"{gi_dir}/GLib-2.0.typelib.tmp", f"{gi_dir}/GLib-2.0.typelib")

    debug("cx_Logging")
    # fixup cx_Logging, required by the service class before we can patch sys.path to find it:
    if os.path.exists(f"{LIB_DIR}/cx_Logging.pyd"):
        os.rename(f"{LIB_DIR}/cx_Logging.pyd", f"{DIST}/cx_Logging.pyd")
    debug("comtypes")
    # fixup cx freeze wrongly including an empty dir:
    gen = f"{LIB_DIR}/comtypes/gen"
    if os.path.exists(gen):
        rmrf(gen)
    debug("gdk loaders")
    lpath = os.path.join(LIB_DIR, "gdk-pixbuf-2.0", "2.10.0", "loaders")
    KEEP_LOADERS = ["jpeg", "png", "xpm", "svg", "wmf", "ico"]
    if not light:
        KEEP_LOADERS += ["pnm", "tiff", "icns"]
    loaders = os.listdir(lpath)
    debug(f"loaders({lpath})={loaders}")
    for filename in loaders:
        if not filename.endswith(".dll") or not any(filename.find(keep) >= 0 for keep in KEEP_LOADERS):
            debug(f"removing {filename!r}")
            os.unlink(os.path.join(lpath, filename))
    debug("gio modules")
    BUNDLE_GIO = ["libproxy", "openssl"]
    if not light:
        BUNDLE_GIO += ["gnutls"]
    os.mkdir(f"{LIB_DIR}/gio")
    os.mkdir(f"{LIB_DIR}/gio/modules")
    for modname in BUNDLE_GIO:
        gio_filename = f"gio/modules/libgio{modname}.dll"
        copyfile(f"{MINGW_PREFIX}/lib/{gio_filename}", f"{LIB_DIR}/{gio_filename}")
    log_command(f"gio-querymodules.exe {LIB_DIR}/gio/modules", "gio-cache.log")
    debug("remove ffmpeg libraries")
    for libname in ("avcodec", "avformat", "avutil", "swscale", "swresample", "zlib1", "xvidcore"):
        find_delete(LIB_DIR, libname)
    debug("move lz4")


def add_numpy(bundle: bool) -> None:
    step(f"numpy: {bundle}")
    lib_numpy = f"{LIB_DIR}/numpy"
    if not bundle:
        debug("removed")
        rmrf(f"{lib_numpy}")
        delete_libs("libopenblas*", "libgfortran*", "libquadmath*")
        return
    debug("moving libraries to lib")
    for libname in ("openblas", "gfortran", "quadmath"):
        for dll in glob(f"{lib_numpy}/core/lib{libname}*.dll"):
            move(dll, LIB_DIR)
    debug("trim tests")


def move_lib(frompath: str, todir: str) -> None:
    topath = os.path.join(todir, os.path.basename(frompath))
    if os.path.exists(topath):
        # should compare that they are the same file!
        debug(f"removing {frompath!r}, already found in {topath!r}")
        os.unlink(frompath)
        return
    debug(f"moving {frompath!r} to {topath!r}")
    move(frompath, topath)


def fixup_gstreamer() -> None:
    step("Fixup GStreamer")
    lib_gst = f"{LIB_DIR}/gstreamer-1.0"
    # these are not modules, so they belong in "lib/":
    for dllname in ("gstreamer*", "gst*-1.0-*", "wavpack*", "*-?"):
        for gstdll in glob(f"{lib_gst}/lib{dllname}.dll"):
            move_lib(gstdll, f"{LIB_DIR}")
    # all the remaining libgst* DLLs are gstreamer elements:
    for gstdll in glob(f"{LIB_DIR}/gst*.dll"):
        move(gstdll, lib_gst)
    # gst-inspect and gst-launch don't need to be at the root:
    for exe in glob(f"{DIST}/gst-*.exe"):
        move_lib(exe, LIB_DIR)
    # these are not needed at all for now:
    for elementname in ("basecamerabinsrc", "photography"):
        for filename in glob(f"{LIB_DIR}/libgst{elementname}*"):
            os.unlink(filename)


def fixup_dlls() -> None:
    step("Fixup DLLs")
    debug("remove dll.a")
    # why is it shipping those files??
    find_delete(DIST, "*dll.a")
    debug("moving most DLLs to lib/")
    # but keep the core DLLs in the root (python, gcc, etc):
    exclude = ("msvcrt", "libpython", "libgcc", "libwinpthread", "pdfium")
    for dll in glob(f"{DIST}/*.dll"):
        if any(dll.find(excl) >= 0 for excl in exclude):
            continue
        move_lib(dll, LIB_DIR)
    debug("fixing cx_Freeze duplication")
    # remove all the pointless cx_Freeze duplication:
    for dll in glob(f"{DIST}/*.dll") + glob(f"{LIB_DIR}/*dll"):
        filename = os.path.basename(dll)
        # delete from any sub-directories:
        find_delete(LIB_DIR, filename, mindepth=2)


def delete_dist_files(*exps: str) -> None:
    for exp in exps:
        matches = glob(f"{DIST}/{exp}")
        if not matches:
            print(f"Warning: glob {exp!r} did not match any files!")
            continue
        for path in matches:
            if os.path.isdir(path):
                debug(f"removing tree at: {path!r}")
                rmtree(path)
            else:
                debug(f"removing {path!r}")
                os.unlink(path)


def delete_libs(*exps: str) -> None:
    debug(f"deleting libraries: {exps}")
    delete_dist_files(*(f"lib/{exp}" for exp in exps))


def delete_dlls(light: bool) -> None:
    step("Deleting unnecessary DLLs")
    delete_libs(
        "libjasper*", "xdg*", "olefile*", "pygtkcompat*", "jaraco*",
        "p11-kit*", "lz4",
    )
    # remove codecs we don't need:
    delete_libs("libx265*", "libjxl*", "libde265*", "libkvazaar*")
    if light:
        # let's keep kerberos / gss libs because clients can use them:
        # "libshishi*", "libgss*"
        delete_libs(
            # no dbus:
            "libdbus*",
            # no AV1:
            "libaom*", "rav1e*", "libdav1d*", "libheif*",
            # no avif:
            "libavif*", "libSvt*",
            # remove h264 encoder:
            "libx264*",
            # should not be needed:
            "libp11-kit*",
            # extra audio codecs (we just keep vorbis and opus):
            "libmp3*", "libwavpack*", "libmpdec*", "libFLAC*", "libmpg123*", "libfaad*", "libfaac*",
        )

        def delgst(*exps: str) -> None:
            gstlibs = tuple(f"gstreamer-1.0/libgst{exp}*" for exp in exps)
            delete_libs(*gstlibs)
        # matching gstreamer modules:
        delgst("flac", "wavpack", "wavenc", "lame", "mpg123", "faac", "faad", "wav")
        # these started causing packaging problems with GStreamer 1.24:
        delgst("isomp4")


def trim_pillow() -> None:
    # remove PIL loaders and modules we don't need:
    step("Removing unnecessary PIL plugins")
    KEEP = (
        "Bmp", "Ico", "Jpeg", "Tiff", "Png", "Ppm", "Xpm", "WebP",
        "Image.py", "ImageChops", "ImageCms", "ImageWin", "ImageChops", "ImageColor", "ImageDraw", "ImageFile.py",
        "ImageFilter", "ImageFont", "ImageGrab", "ImageMode", "ImageOps", "ImagePalette", "ImagePath", "ImageSequence",
        "ImageStat", "ImageTransform",
    )
    NO_KEEP = ("Jpeg2K", )
    kept = []
    removed = []
    for filename in glob(f"{LIB_DIR}/PIL/*Image*"):
        infoname = os.path.splitext(os.path.basename(filename))[0]
        if any(filename.find(keep) >= 0 for keep in KEEP) and not any(filename.find(nokeep) >= 0 for nokeep in NO_KEEP):
            kept.append(infoname)
            continue
        removed.append(infoname)
        os.unlink(filename)
    debug(f"removed: {csv(removed)}")
    debug(f"kept: {csv(kept)}")


def trim_python_libs() -> None:
    step("Removing unnecessary Python modules")
    # remove test bits we don't need:
    delete_libs(
        "backports",
        "importlib_resources/compat",
        "importlib_resources/tests",
        # no need for headers:
        # "cairo/include"
    )
    step("Removing unnecessary files")
    for ftype in (
        # no runtime type checks:
        "py.typed",
        # remove source:
        "*.bak",
        "*.orig",
        "*.pyx",
        "*.c",
        "*.cpp",
        "*.m",
        "constants.txt",
        "*.h",
        "*.html",
        "*.pxd",
        "*.cu",
    ):
        find_delete(DIST, ftype)


def fixup_zeroconf() -> None:
    lib_zeroconf = f"{LIB_DIR}/zeroconf"
    rmrf(lib_zeroconf)
    # workaround for zeroconf - just copy it wholesale
    # since I have no idea why cx_Freeze struggles with it:
    zc = find_spec("zeroconf")
    if not zc:
        print("Warning: zeroconf not found for Python %s" % sys.version)
        return
    zeroconf_dir = os.path.dirname(zc.origin)
    debug(f"adding zeroconf from {zeroconf_dir!r} to {lib_zeroconf!r}")
    copytree(zeroconf_dir, lib_zeroconf)


def rm_empty_dir(dirpath: str) -> None:
    cmd = ["find", dirpath, "-type", "d", "-empty"]
    output = check_output(command_args(cmd))
    for path in output.splitlines():
        os.rmdir(path)


def rm_empty_dirs() -> None:
    step("Removing empty directories")
    for _ in range(3):
        rm_empty_dir(DIST)


def zip_modules(light: bool) -> None:
    step("zipping up some Python modules")
    # these modules contain native code or data files,
    # so they will require special treatment:
    # xpra numpy cryptography PIL nacl cffi gtk gobject glib aioquic pylsqpack > /dev/null
    ZIPPED = [
        "OpenGL", "encodings", "future", "paramiko", "html",
        "pyasn1", "asn1crypto", "async_timeout",
        "OpenSSL", "keyring",
        "ifaddr", "pyaes", "service_identity",
        "re", "platformdirs", "attr", "setproctitle", "pyvda", "zipp",
        "distutils", "comtypes", "email", "multiprocessing", "packaging",
        "pkg_resources", "pycparser", "idna", "ctypes", "json",
        "http", "enum", "winreg", "copyreg", "_thread", "_dummythread",
        "builtins", "importlib",
        "logging", "queue", "urllib", "xml", "xmlrpc", "pyasn1_modules",
        "concurrent", "collections",
    ]
    if not light:
        ZIPPED += EXTRA_PYTHON_MODULES
    log_command(["zip", "--move", "-ur", "library.zip"] + ZIPPED, "zip.log", cwd=LIB_DIR)


def setup_share(light: bool) -> None:
    step("Deleting unnecessary `share/` files")
    delete_dist_files(
        "share/xml",
        "share/glib-2.0/codegen",
        "share/glib-2.0/gdb",
        "share/glib-2.0/gettext",
        "share/locale",
        "share/gstreamer-1.0",
        "share/gst-plugins-base",
        "share/p11-kit",
    )
    if light:
        # remove extra bits that take up a lot of space:
        delete_dist_files(
            "share/icons/Adwaita/cursors",
            "share/fonts/gsfonts",
            "share/fonts/adobe*",
            "share/fonts/cantarell",
        )
    step("Removing empty icon directories")
    # remove empty icon directories
    for _ in range(4):
        rm_empty_dir(f"{DIST}/share/icons")


def add_manifests(light: bool) -> None:
    step("Adding EXE manifests")
    EXES = [
        "Bug_Report", "Xpra-Launcher", "Xpra", "Xpra_cmd", "Configure", "PDFIUM_Print",
    ]
    if not light:
        EXES += [
            "Xpra-Shadow", "Xpra-Proxy_cmd", "Xpra-Proxy",
            "Python_exec_cmd", "Python_exec_gui",
            "Xpra-Launcher-Debug",
            "Config_info",
            "Python_execfile_cmd", "Python_execfile_gui",
            "Print", "NVidia_info", "NvFBC_capture", "CUDA_info",
            "Auth_Dialog",
            "OpenGL_check",
            "GTK_info", "NativeGUI_info", "Screenshot",
            "Version_info", "Network_info", "Keymap_info", "Keyboard_info",
            "SystemTray_Test", "U2F_Tool",
            "SQLite_auth_tool", "SQL_auth_tool",
            "Encoding_info", "Path_info", "Feature_info", "NativeGUI_info",
            "Xpra_Audio",
            "GStreamer_info", "Audio_Devices",
            "gst-launch-1.0", "gst-inspect-1.0",
            "Webcam_info", "Webcam_Test",
            "System-Auth-Test", "LDAP-Auth-Test", "LDAP3-Auth-Test",
            "System-Logon-Test", "Events_Test", "GTK_Keyboard_Test",
        ]
    with open('packaging/MSWindows/exe.manifest', 'r') as file:
        xml_file_content = file.read()
    zero_padded_version = version_info.padded
    xml_file_content = xml_file_content.replace("XPRA_ZERO_PADDED_VERSION", zero_padded_version)
    with open('packaging/MSWindows/exe.manifest.tmp', 'w') as file:
        file.write(xml_file_content)
    for exe in EXES:
        copyfile("packaging/MSWindows/exe.manifest.tmp", f"{DIST}/{exe}.exe.manifest")
    os.remove('packaging/MSWindows/exe.manifest.tmp')


def gen_caches() -> None:
    step("Generating gdk pixbuf loaders cache")
    cmd = ["gdk-pixbuf-query-loaders.exe"]
    for loader in glob(f"{DIST}/lib/gdk-pixbuf-2.0/2.10.0/loaders/*"):
        cmd.append(loader.split(f"{DIST}/", 1)[1])
    with Popen(cmd, cwd=os.path.abspath(DIST), stdout=PIPE, text=True) as proc:
        cache, err = proc.communicate(None)
    if proc.returncode != 0:
        raise RuntimeError(f"gdk-pixbuf-query-loaders.exe failed and returned {proc.returncode}: {err}"
                           " - you may need to run `chcp.com 65001`")
    # replace absolute paths:
    cache = re.sub(r'".*xpra/dist/lib/', '"lib/', cache)
    with open(f"{LIB_DIR}/gdk-pixbuf-2.0/2.10.0/loaders.cache", "w") as cache_file:
        cache_file.write(cache)
    step("Generating icons and theme cache")
    for itheme in glob(f"{DIST}/share/icons/*"):
        log_command(["gtk-update-icon-cache.exe", "-t", "-i", itheme], "icon-cache.log")


def bundle_manual() -> None:
    step("Generating HTML Manual Page")
    manual = os.path.join(DIST, "manual.html")
    delfile(manual)
    with open("fs/share/man/man1/xpra.1", "rb") as f:
        man = f.read()
    proc = Popen(["groff", "-mandoc", "-Thtml"], stdin=PIPE, stdout=PIPE, stderr=PIPE)
    out, err = proc.communicate(man)
    if proc.returncode != 0:
        raise RuntimeError(f"groff failed and returned {proc.returncode}: {err!r}")
    debug(f"groff warnings: {err!r}")
    with open(manual, "wb") as manual_file:
        manual_file.write(out)


def bundle_html5() -> None:
    step("Installing the HTML5 client")
    www = os.path.join(os.path.abspath("."), DIST, "www")
    if not os.path.exists(www):
        os.mkdir(www)
    html5 = os.path.join(os.path.abspath("."), "xpra-html5")
    debug(f"running html5 install step in {html5!r}")
    log_command([PYTHON, "./setup.py", "install", www], "html5.log", cwd=html5)


def bundle_putty() -> None:
    step("Bundling TortoisePlink")
    tortoiseplink = find_command("TortoisePlink", "TORTOISEPLINK",
                                 f"{PROGRAMFILES}\\TortoiseSVN\\bin\\TortoisePlink.exe")
    copyfile(tortoiseplink, f"{DIST}/Plink.exe")
    for dll in ("vcruntime140.dll", "msvcp140.dll", "vcruntime140_1.dll"):
        copyfile(f"{SYSTEM32}/{dll}", f"{DIST}/{dll}")


def bundle_dlls(*expr: str) -> None:
    for exp in expr:
        matches = glob(f"{exp}.dll")
        if not matches:
            print(f"Warning: no dll matching {exp!r}")
            continue
        for match in matches:
            name = os.path.basename(match)
            copyfile(match, f"{DIST}/{name}")


def bundle_openssh() -> None:
    step("Bundling OpenSSH")
    for exe_name in ("ssh", "sshpass", "ssh-keygen"):
        exe = which(exe_name)
        if not exe:
            raise RuntimeError(f"{exe_name!r} not found!")
        copyfile(exe, f"{DIST}/{exe_name}.exe")
    bin_dir = os.path.dirname(which("ssh"))
    debug(f"looking for msys DLLs in {bin_dir!r}")
    msys_dlls = tuple(
        f"{bin_dir}/msys-{dllname}*" for dllname in (
            "2.0", "gcc_s", "crypto", "z", "gssapi", "asn1", "com_err", "roken",
            "crypt", "heimntlm", "krb5", "heimbase", "wind", "hx509", "hcrypto", "sqlite3",
        )
    )
    bundle_dlls(*msys_dlls)


def bundle_openssl() -> None:
    step("Bundling OpenSSL")
    copyfile(f"{MINGW_PREFIX}/bin/openssl.exe", f"{DIST}/openssl.exe")
    ssl_dir = f"{DIST}/etc/ssl"
    if not os.path.exists(ssl_dir):
        os.mkdir(ssl_dir)
    copyfile(f"{MINGW_PREFIX}/etc/ssl/openssl.cnf", f"{ssl_dir}/openssl.cnf")
    # we need those libraries at the top level:
    bundle_dlls(f"{LIB_DIR}/libssl-*", f"{LIB_DIR}/libcrypto-*")


def bundle_paxec() -> None:
    step("Bundling paexec")
    copyfile(f"{MINGW_PREFIX}/bin/paexec.exe", f"{DIST}/paexec.exe")


def bundle_desktop_logon() -> None:
    step("Bundling desktop_logon")
    dl_dlls = tuple(f"{MINGW_PREFIX}/bin/{dll}" for dll in ("AxMSTSCLib", "MSTSCLib", "DesktopLogon"))
    bundle_dlls(*dl_dlls)


def add_cuda(enabled: bool) -> None:
    step(f"cuda: {enabled}")
    if not enabled:
        delete_libs("pycuda*")
        find_delete(DIST, "pycuda*")
        find_delete(DIST, "libnv*")
        find_delete(DIST, "cuda.conf")
        delete_libs("curand*")
        cuda_dir = os.path.join(LIB_DIR, "cuda")
        if os.path.exists(cuda_dir):
            rmtree(cuda_dir)
        return
    # pycuda wants a CUDA_PATH with "/bin" in it:
    if not os.path.exists(f"{DIST}/bin"):
        os.mkdir(f"{DIST}/bin")
    # keep the cuda bits at the root:
    for nvdll in glob(f"{LIB_DIR}/libnv*.dll"):
        move_lib(nvdll, DIST)


def rec_options(args) -> None:
    info = dict((k, getattr(args, k)) for k in dir(args) if not k.startswith("_"))
    with open(BUILD_INFO, "a") as f:
        f.write(f"\n\nBUILD_OPTIONS={info!r}\n")


def find_source(path: str) -> str:
    basename = os.path.basename(path)
    if path.endswith(".exe"):
        filename = which(basename)
        if filename:
            return filename
    # try "$MINGW_PREFIX/bin/$basename" and "$MINGW_PREFIX/$path":
    for filename in (os.path.join(MINGW_PREFIX, "bin", basename), os.path.join(MINGW_PREFIX, path)):
        if os.path.exists(filename):
            return filename
    # try %PATH%:
    for bpath in os.environ.get("PATH", "").split(os.path.pathsep):
        filename = os.path.join(bpath, basename)
        if os.path.exists(filename):
            return filename
    return ""


def get_command_info(*args: str) -> dict[str, str]:
    cmd = command_args(list(args))
    r, output = getstatusoutput(cmd)
    if r:
        debug(f"{args} failed and returned {r}")
        return {}
    props: dict[str, str] = {}
    for line in output.split("\n"):
        parts = line.split(":", 1)
        if len(parts) == 2:
            props[parts[0].strip()] = parts[1].strip()
    return props


def get_pacman_package_info(name: str) -> dict[str, str]:
    return get_command_info("pacman", "-Qi", name)


def get_pip_package_info(name: str) -> dict[str, str]:
    return get_command_info("pip", "show", name)


def get_package(path: str) -> tuple[str, str]:
    filename = find_source(path)
    if not filename:
        debug(f"failed to locate source path for {path!r}")
        return "", ""
    cmd = command_args(["pacman", "-Qo", filename])
    r, output = getstatusoutput(cmd)
    if r:
        debug(f"pacman failed for {filename!r} and returned {r}")
        return "", ""
    # ie: "/usr/bin/msys-2.0.dll is owned by msys2-runtime 3.5.4-2" ->
    #     ["/usr/bin/msys-2.0.dll", "msys2-runtime 3.5.4-2"]
    parts = output.split("\n")[0].split("is owned by ")
    if len(parts) != 2:
        debug(f"unable to parse pacman output: {output!r}")
        return "", ""
    # ie: "msys2-runtime 3.5.4-2" -> ["msys2-runtime", "3.5.4-2"]
    package, version = parts[1].split(" ", 1)
    return package, version


def get_py_lib_dirs() -> list[str]:
    # python modules:
    py_lib_dirs: list[str] = []
    for k in ("stdlib", "platstdlib", "purelib", "platlib", ):
        path = get_path(k)
        if os.path.exists(path) and path not in py_lib_dirs:
            py_lib_dirs.append(path)
            for extra in ("lib-dynload", "win32", "pywin32_system32"):
                epath = os.path.join(path, extra)
                if os.path.exists(epath) and epath not in py_lib_dirs:
                    py_lib_dirs.append(epath)
    debug(f"using python lib directories: {py_lib_dirs!r}")
    return py_lib_dirs


def pip_sbom_rec(prefix: str, filename: str) -> dict[str, Any]:
    pip_rec = get_pip_package_info(filename)
    if not pip_rec:
        return {}
    rec = {
        "package": "python-" + pip_rec["Name"],
        "version": pip_rec["Version"],
        "pip": True,
    }
    dist_path = os.path.join(prefix, filename)
    add_size_checksum(rec, dist_path)
    return rec


def sbom_rec(path: str) -> dict[str, Any]:
    package, version = get_package(path)
    if not (package and version):
        return {}
    rec = {
        "package": package,
        "version": version,
    }
    dist_path = os.path.join(DIST, path)
    add_size_checksum(rec, dist_path)
    return rec


def add_size_checksum(rec: dict[str, Any], dist_path: str) -> None:
    if os.path.isdir(dist_path):
        rec["size"] = du(dist_path)
    else:
        rec["size"] = os.stat(dist_path).st_size
        with open(dist_path, "rb") as f:
            data = f.read()
        rec["checksum"] = hashlib.sha256(data).hexdigest()


def find_glob_paths(dirname: str, glob_str: str) -> list[str]:
    cmd = command_args(["find", dirname, "-name", "'"+glob_str+"'"])
    r, output = getstatusoutput(cmd)
    assert r == 0, f"{cmd!r} failed and returned {r}"
    return output.splitlines()


def rec_sbom() -> None:
    step("Recording SBOM")
    sbom: dict[str, dict[str, Any]] = {}
    py_lib_dirs: list[str] = get_py_lib_dirs()

    def find_prefixed_sbom_rec(filename: str, prefixes: list[str]) -> dict[str, Any]:
        for prefix in prefixes:
            src_path = os.path.join(prefix, filename)
            if os.path.exists(src_path):
                rec = sbom_rec(src_path) or pip_sbom_rec(prefix, filename)
                if rec:
                    package = rec["package"]
                    version = rec["version"]
                    debug(f" * {filename!r}: {package!r}, {version!r}")
                    return rec
        print(f"Warning: unknown source for filename {filename!r}, tried {prefixes}")
        return {}

    def rec_py_lib(path: str) -> None:
        assert path.startswith("lib/"), f"invalid path {path!r}"
        name = os.path.basename(path)
        rec = find_prefixed_sbom_rec(name, py_lib_dirs)
        if rec:
            sbom[f"lib/{path}"] = rec

    def rec_pyqt_lib(path: str) -> None:
        assert path.startswith("lib/PyQt6")
        share_dirs = os.environ.get("XDG_DATA_DIRS", f"{MINGW_PREFIX}/share").split(os.path.pathsep)
        dirs = []
        for share_dir in share_dirs:
            dirs += glob(f"{share_dir}/qt6/plugins/*")
        name = os.path.basename(path)
        rec = find_prefixed_sbom_rec(name, dirs)
        if rec:
            sbom[path] = rec

    def rec_cuda(path: str) -> None:
        if not os.path.exists("CUDA") or not os.path.isdir("CUDA"):
            raise RuntimeError("CUDA not found!")
        version_json = os.path.join("CUDA", "version.json")
        if not os.path.exists(version_json):
            raise RuntimeError(f"{version_json!r} does not exist")
        with open(version_json, "r") as f:
            version_data = json.loads(f.read())
        cuda_version = version_data["cuda"]["version"]
        sbom[path] = {
            "package": "cuda",
            "version": cuda_version,
        }

    globbed_paths = find_glob_paths(DIST, "*.dll") + find_glob_paths(LIB_DIR, "*.exe")
    debug(f"adding DLLs and EXEs: {globbed_paths}")
    for globbed_path in globbed_paths:
        path = globbed_path[len(DIST)+1:]
        if path.startswith("lib/PyQt6"):
            rec_pyqt_lib(path)
        elif path.startswith("lib/python"):
            # some need to be resolved in the python library paths:
            rec_py_lib(path)
        elif path.startswith("lib/curand"):
            rec_cuda(path)
        elif path in ("AxMSTSCLib.dll", "MSTSCLib.dll", "vcruntime140.dll", "vcruntime140_1.dll", "msvcp140.dll"):
            debug(f"ignoring {path!r} (.NET SDK)")
        elif path in ("DesktopLogon.dll", ):
            debug(f"ignoring {path!r} (one of ours)")
        else:
            rec = sbom_rec(path)
            if rec:
                package = rec["package"]
                version = rec["version"]
                debug(f" * {path!r}: {package!r}, {version!r}")
                sbom[path] = rec
            else:
                print(f"Warning: no package data found for {path!r}")

    # python modules:
    debug("adding python modules")
    SKIP_DIRS = (
        "xpra", "tlb",
        "gi", "gio", "pkcs11", "girepository-1.0", "gstreamer-1.0", "gtk-3.0", "gdk-pixbuf-2.0",
    )
    for child in os.listdir(LIB_DIR):
        if child in SKIP_DIRS or child.endswith(".dll"):
            continue
        path = os.path.join(LIB_DIR, child)
        if os.path.isdir(path) or path.endswith(".py") or path.endswith(".pyd"):
            rec_py_lib(os.path.join("lib", child))
    # add this one by hand because cx_Freeze hides it in `library.zip`,
    # and we don't want to start unpacking a large ZIP file and converting .pyc to .py
    # just for one filename:
    rec_py_lib(os.path.join("lib", "decorator.py"))

    # summary: list of packages
    packages = tuple(sorted(set(rec["package"] for rec in sbom.values())))
    debug(f"adding package info for {packages}")
    packages_info: dict[str, tuple] = {}
    for package_name in packages:
        # shorten the package name:
        package = package_name
        if package.startswith(PACKAGE_PREFIX):
            package = package[len(PACKAGE_PREFIX):]
        elif os.path.basename(package).startswith(MSYS_DLL_PREFIX):
            # ie: "msys-com_err-1.dll"
            pass  # keep it as it is
        elif package.startswith(MSYS2_PACKAGE_PREFIX):
            # ie: "msys2-runtime"
            pass  # keep it as it is
        else:
            debug(f"unexpected package prefix: {package!r}")
        # keep only the keys relevant to the sbom:
        info = get_pacman_package_info(package_name) or get_pip_package_info(package_name)
        exported_info = {}
        for key in (
            "Name", "Version", "Description",
            "Architecture", "URL", "Licenses", "Depends On", "Required By", "Provides",
        ):
            value = info.get(key)
            if str(value) == "None":
                continue
            if key in ("Depends On", "Required By", "Provides"):
                value = [x.strip() for x in str(value).split(" ") if x.strip()]
            exported_info[key] = value
        packages_info[package] = exported_info

    debug(f"recording sbom data: {len(sbom)} paths, {len(packages)} packages")
    with open(BUILD_INFO, "a") as f:
        f.write(f"\n# {len(sbom)} SBOM path entries:\n")
        f.write(f"sbom={sbom!r}\n")
        f.write(f"\n# {len(packages_info)} packages:\n")
        f.write(f"packages={packages_info!r}\n")
    # also replace it in the target directory:
    find_delete(LIB_DIR, os.path.basename(BUILD_INFO))
    copyfile(BUILD_INFO, f"{LIB_DIR}/{BUILD_INFO}")


def export_sbom() -> None:
    WIN_PYTHON = "C:\\Program Files\\Python312\\python.exe"
    SBOM_SCRIPT = "packaging\\MSWindows\\cyclonedx_sbom.py"
    output = f"{DIST}/{SBOM_JSON}"
    delfile(output)
    env = os.environ.copy()
    env["PYTHONPATH"] = "."
    log_command([WIN_PYTHON, SBOM_SCRIPT, output], "export-sbom.log", env=env)
    # make a copy that can be distributed separately:
    SBOM_FILENAME = f"Xpra{version_info.extra}{version_info.arch_info}_{version_info.full_string}.json"
    delfile(SBOM_FILENAME)
    copyfile(output, SBOM_FILENAME)


def verpatch() -> None:
    EXCLUDE = ("plink", "openssh", "openssl", "paexec")

    def run_verpatch(filename: str, descr: str) -> None:
        log_command(["verpatch", filename,
                     "/s", "desc", descr,
                     "/va", version_info.padded,
                     "/s", "company", "xpra.org",
                     "/s", "copyright", "(c) xpra.org 2024",
                     "/s", "product", "xpra",
                     "/pv", version_info.padded,
                     ], "verpatch.log")

    for exe in glob(f"{DIST}/*.exe"):
        if any(exe.lower().find(excl) >= 0 for excl in EXCLUDE):
            continue
        exe_name = os.path.basename(exe)
        if exe_name in ("Xpra_cmd.exe", "Xpra.exe", "Xpra-Proxy.exe"):
            # handled separately below
            continue
        assert exe_name.endswith(".exe")
        tool_name = exe_name[:-3].replace("Xpra_", "").replace("_", " ").replace("-", " ")
        run_verpatch(exe, f"Xpra {tool_name}")
    run_verpatch(f"{DIST}/Xpra_cmd.exe", "Xpra command line")
    run_verpatch(f"{DIST}/Xpra.exe", "Xpra")
    if os.path.exists(f"{DIST}/Xpra-Proxy.exe"):
        run_verpatch(f"{DIST}/Xpra-Proxy.exe", "Xpra Proxy Server")


################################################################################
# packaging: ZIP / EXE / MSI

def create_zip() -> None:
    step("Creating ZIP file:")
    ZIP_DIR = f"Xpra{version_info.extra}{version_info.arch_info}_{version_info.full_string}"
    ZIP_FILENAME = f"{ZIP_DIR}.zip"
    if os.path.exists(ZIP_DIR):
        rmrf(ZIP_DIR)
    delfile(ZIP_FILENAME)
    copytree(DIST, ZIP_DIR)
    log_command(["zip", "-9mr", ZIP_FILENAME, ZIP_DIR], "zip.log")
    size = du(ZIP_FILENAME) // 1024 // 1024
    print(f"{ZIP_FILENAME}: {size}MB")


def create_exe(args) -> str:
    step("Creating the EXE installer using InnoSetup")
    innosetup = find_command("innosetup", "INNOSETUP",
                             f"{PROGRAMFILES}\\Inno Setup 6\\ISCC.exe",
                             f"{PROGRAMFILES_X86}\\Inno Setup 6\\ISCC.exe",
                             )
    SETUP_EXE = f"{DIST}/Xpra_Setup.exe"
    INSTALLER_FILENAME = f"Xpra{version_info.extra}{version_info.arch_info}_Setup_{version_info.full_string}.exe"
    XPRA_ISS = "xpra.iss"
    INNOSETUP_LOG = "innosetup.log"
    for filename in (XPRA_ISS, INNOSETUP_LOG, INSTALLER_FILENAME, SETUP_EXE):
        delfile(filename)
    with open("packaging/MSWindows/xpra.iss", "r") as f:
        contents = f.readlines()
    lines = []
    subs = {
        "AppId": "Xpra_is1",
        "AppName": f"Xpra {version_info.string}",
        "UninstallDisplayName": f"Xpra {version_info.string}",
        "AppVersion": version_info.full_string,
    }
    for line in contents:
        if line.startswith("    PostInstall()") and args.light:
            # don't run post-install openssl:
            line = "Log('skipped post-install');"
        elif line.find("Xpra Shadow Server") >= 0 and args.light:
            # no shadow server in light builds
            continue
        elif line.find("Command Manual") >= 0 and not args.docs:
            # remove link to the manual:
            continue
        if line.find("=") > 0:
            parts = line.split("=", 1)
            if parts[0] in subs:
                line = parts[0] + "=" + subs[parts[0]]+"\n"
        lines.append(line)
    with open(XPRA_ISS, "w") as f:
        f.writelines(lines)

    log_command([innosetup, XPRA_ISS], INNOSETUP_LOG)
    os.unlink(XPRA_ISS)

    os.rename(SETUP_EXE, INSTALLER_FILENAME)
    size = du(INSTALLER_FILENAME) // 1024 // 1024
    print(f"{INSTALLER_FILENAME}: {size}MB")
    return INSTALLER_FILENAME


def sign_file(filename: str) -> None:
    log_command(["signtool.exe", "sign", "/fd", "SHA256", "/v", "/f", KEY_FILE, "/t", TIMESTAMP_SERVER, filename], "signtool.log")


def create_msi(exe: str) -> str:
    msiwrapper = find_command("msiwrapper", "MSIWRAPPER",
                              f"{PROGRAMFILES}\\MSI Wrapper\\MsiWrapper.exe",
                              f"{PROGRAMFILES_X86}\\MSI Wrapper\\MsiWrapper.exe")
    MSI_FILENAME = f"Xpra{version_info.extra}{version_info.arch_info}_{version_info.full_string}.msi"
    # search and replace in the template file:
    subs: dict[str, str] = {
        "CWD": os.getcwd(),
        "INPUT": exe,
        "OUTPUT": MSI_FILENAME,
        "ZERO_PADDED_VERSION": version_info.padded,
        "FULL_VERSION": version_info.full_string,
    }
    with open("packaging\\MSWindows\\msi.xml", "r") as template:
        msi_data = template.read()
    for varname, value in subs.items():
        msi_data = msi_data.replace(f"${varname}", value)
    MSI_XML = "msi.xml"
    with open(MSI_XML, "w") as f:
        f.write(msi_data)
    log_command([msiwrapper, MSI_XML], "msiwrapper.log")
    size = du(MSI_FILENAME) // 1024 // 1024
    print(f"{MSI_FILENAME}: {size}MB")
    return MSI_FILENAME


def build(args) -> None:
    set_version_info(args.light)
    if args.html5:
        check_html5()
    if args.sign:
        check_signtool()

    if args.clean:
        clean()
    if args.service:
        build_service()
    if args.cuda:
        build_cuda_kernels()
    if args.build:
        build_ext(args)
    if args.tests:
        run_tests()
    if args.install:
        add_manifests(args.light)
        install_exe(args)

    if args.fixups:
        fixups(args.light)
        fixup_gstreamer()
        fixup_dlls()
        delete_dlls(args.light)
        trim_python_libs()
        trim_pillow()
        fixup_zeroconf()
        if args.light:
            delete_libs(*EXTRA_PYTHON_MODULES)
        rm_empty_dirs()

    add_cuda(args.cuda)
    add_numpy(args.numpy)

    setup_share(args.light)
    gen_caches()

    if args.docs:
        bundle_manual()
        install_docs()
    if args.html5:
        bundle_html5()
    if args.putty:
        bundle_putty()
    if args.openssh:
        bundle_openssh()
    if args.openssl:
        bundle_openssl()
    if args.paexec:
        bundle_paxec()
    if args.desktop_logon:
        bundle_desktop_logon()
    rec_options(args)
    if args.sbom:
        rec_sbom()
        export_sbom()

    if args.zip_modules:
        zip_modules(args.light)

    if args.verpatch:
        verpatch()

    size = du(DIST) // 1024 // 1024
    print(f"installed size: {size}MB")
    if args.zip:
        create_zip()
    if args.exe:
        exe = create_exe(args)
        if not os.path.exists(exe):
            raise RuntimeError(f"failed to create EXE installer {exe!r}")
        if args.sign:
            step("Signing EXE")
            sign_file(exe)
        if args.run:
            step(f"Running the new EXE installer {exe!r}")
            os.system(exe)
        if args.msi:
            msi = create_msi(exe)
            if not os.path.exists(msi):
                raise RuntimeError("failed to create msi")
            if args.sign:
                step("Signing MSI")
                sign_file(msi)


def main(argv) -> None:
    args, unknownargs = parse_command_line(argv)
    if len(unknownargs) > 1:
        if len(unknownargs) != 2:
            raise ValueError(f"too many arguments: {unknownargs!r}")
        arg = unknownargs[1]
        if arg == "sbom":
            rec_sbom()
        elif arg == "export-sbom":
            load_version_info(args.light)
            export_sbom()
        elif arg == "gen-caches":
            gen_caches()
        elif arg == "zip":
            create_zip()
        elif arg == "exe":
            create_exe(args)
        else:
            raise ValueError(f"unknown argument {arg!r}")
    else:
        build(args)


if __name__ == "__main__":
    main(sys.argv)



# This file is part of Xpra.
# Copyright (C) 2021 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import struct
from threading import RLock
from collections.abc import Callable

from xpra.os_util import gi_import
from xpra.net.rfb.protocol import RFBProtocol
from xpra.net.rfb.const import RFBEncoding, RFBClientMessage, RFBAuth, CLIENT_INIT, AUTH_STR, RFB_KEYS
from xpra.util.str_fn import csv, repr_ellipsized, hexstr
from xpra.log import Logger

GLib = gi_import("GLib")

log = Logger("network", "protocol", "rfb")

WID = 1


def check_wid(wid) -> bool:
    if wid != WID:
        log("ignoring pointer movement outside the VNC window")
        return False
    return True


class RFBClientProtocol(RFBProtocol):

    def __init__(self, conn, process_packet_cb, next_packet):
        self.next_packet = next_packet
        self.rectangles = 0
        self.position = 0, 0
        # translate xpra packets into rfb packets:
        self._rfb_converters: dict[str, Callable] = {
            "pointer-position": self.send_pointer_position,
            "button-action": self.send_button_action,
            "key-action": self.send_key_action,
            "configure-window": self.track_window,
        }
        self.send_lock = RLock()
        super().__init__(conn, process_packet_cb)

    def source_has_more(self) -> None:
        log("source_has_more()")
        if not self.send_lock.acquire(False):
            return
        try:
            while True:
                pdata = self.next_packet()
                packet = pdata[0]
                start_send_cb = pdata[1]
                end_send_cb = pdata[2]
                has_more = pdata[5]
                if start_send_cb:
                    start_send_cb()
                log("packet: %s", packet[0])
                handler = self._rfb_converters.get(packet[0])
                if handler:
                    handler(packet)
                if end_send_cb:
                    end_send_cb()
                if not has_more:
                    break
        finally:
            self.send_lock.release()

    def send_pointer_position(self, packet) -> None:
        log("send_pointer_position(%s)", packet)
        # ['pointer-position', 1, (3348, 582), ['mod2'], []]
        if not check_wid(packet[1]):
            return
        x, y = packet[2]
        # modifiers = packet[3]
        buttons = packet[4]
        button_mask = 0
        for i in range(8):
            if i + 1 in buttons:
                button_mask |= 2 ** i
        self.do_send_pointer_event(button_mask, x, y)

    def send_button_action(self, packet) -> None:
        log("send_button_action(%s)", packet)
        if not check_wid(packet[1]):
            return
        # ["button-action", wid, button, pressed, (x, y), modifiers, buttons]
        # ['button-action', 1, 1, False, (2768, 257), ['mod2'], [1]]
        button = packet[2]
        pressed = packet[3]
        x, y = packet[4]
        button_mask = 0
        if pressed:
            button_mask |= 2 ** button
        if len(packet) >= 7:
            buttons = packet[6]
            for i in range(8):
                if i + 1 in buttons:
                    button_mask |= 2 ** i
        self.do_send_pointer_event(button_mask, x, y)

    def do_send_pointer_event(self, button_mask, x, y) -> None:
        # adjust for window position:
        wx, wy = self.position
        self.send_struct(b"!BBHH", RFBClientMessage.PointerEvent, button_mask, x - wx, y - wy)

    def send_key_action(self, packet) -> None:
        log("send_key_action(%s)", packet)
        if not check_wid(packet[1]):
            return
        # ["key-action", "wid", "keyname", "pressed", "modifiers", "keyval", "string", "keycode", "group"]
        keyname = packet[2]
        if len(keyname) == 1:
            keysym = ord(keyname[0])
        else:
            keysym = RFB_KEYS.get(keyname.lower())
        if not keysym:
            log("no keysym found for %s", packet[2:])
            return
        pressed = packet[3]
        self.send_struct(b"!BBHI", RFBClientMessage.KeyEvent, pressed, 0, keysym)

    def track_window(self, packet) -> None:
        log("track_window(%s)", packet)
        if not check_wid(packet[1]):
            return
        self.position = packet[2], packet[3]
        log("window offset: %s", self.position)
        # ["configure-window", self.wid, sx, sy, sw, sh, props, self._resize_counter, state, skip_geometry]
        # ['configure-window', 1, 0, 37, 1280, 1024,
        #    {'encodings.rgb_formats': ['BGRA', 'BGRX', 'RGBA', 'RGBX', 'BGR', 'RGB', 'r210', 'BGR565'],
        #     'encoding.transparency': False,
        #     'encoding.full_csc_modes': {
        #         'h264': ['ARGB', 'BGRA', 'BGRX', 'GBRP', 'GBRP10', 'GBRP9LE',
        #                  'RGB', 'XRGB', 'YUV420P', 'YUV422P', 'YUV444P', 'YUV444P10', 'r210'],
        #         'vp8': ['YUV420P'], 'h265': ['BGRX', 'GBRP', 'GBRP10', 'GBRP9LE',
        #                  'RGB', 'XRGB', 'YUV420P', 'YUV422P', 'YUV444P', 'YUV444P10', 'r210'],
        #         'mpeg4': ['YUV420P'],
        #         'mpeg1': ['YUV420P'],
        #         'mpeg2': ['YUV420P'],
        #         'vp9': ['YUV420P', 'YUV444P', 'YUV444P10'],
        #         'webp': ['BGRA', 'BGRX', 'RGBA', 'RGBX']
        #     },
        #     'encoding.send-window-size': True,
        #     'encoding.render-size': (1280, 1024),
        #     'encoding.scrolling': True},
        #     0, {}, False, 1, (4582, 1055), ['mod2']))

    def handshake_complete(self) -> None:
        log.info("RFB connected to %s", self._conn.target)
        self._packet_parser = self._parse_security_handshake
        self.send_protocol_handshake()

    def _parse_security_handshake(self, packet) -> int:
        log("parse_security_handshake(%s)", hexstr(packet))
        n = struct.unpack(b"B", packet[:1])[0]
        if n == 0:
            self._internal_error("cannot parse security handshake " + hexstr(packet))
            return 0
        security_types = struct.unpack(b"B" * n, packet[1:])
        st = []
        for v in security_types:
            try:
                v = RFBAuth(v)
            except ValueError:
                pass
            st.append(v)
        log("parse_security_handshake(%s) security_types=%s", hexstr(packet), [AUTH_STR.get(v, v) for v in st])
        if not st or RFBAuth.NONE in st:
            auth_type = RFBAuth.NONE
            # go straight to the result:
            self._packet_parser = self._parse_security_result
        elif RFBAuth.VNC in st:
            auth_type = RFBAuth.VNC
            self._packet_parser = self._parse_vnc_security_challenge
        else:
            self._internal_error("no supported security types in %r" % csv(AUTH_STR.get(v, v) for v in st))
            return 0
        self.send_struct(b"B", auth_type)
        return 1 + n

    def _parse_vnc_security_challenge(self, packet) -> int:
        if len(packet) < 16:
            return 0
        challenge = packet[:16]
        log("parse_vnc_security_challenge(%s)", packet)
        auth_caps = {}
        # this will end up calling send_challenge_reply() with the response,
        # the password will be obtained from the client's challenge handlers,
        # which may prompt the user.
        # (see client base for details)
        self._process_packet_cb(self, ["challenge", challenge, auth_caps, "des", "none"])
        return 16

    def send_challenge_reply(self, challenge_response) -> None:
        log("send_challenge_reply(%s)", challenge_response)
        self._packet_parser = self._parse_security_result
        import binascii  # pylint: disable=import-outside-toplevel
        self.send(binascii.unhexlify(challenge_response))

    def _parse_security_result(self, packet) -> int:
        if len(packet) < 4:
            return 0
        r = struct.unpack(b"I", packet[:4])[0]
        if r != 0:
            self._internal_error(f"authentication denied, server returned {r}")
            return 0
        log("parse_security_result(%s) success", hexstr(packet))
        self._packet_parser = self._parse_client_init
        share = False
        self.send_struct(b"B", bool(share))
        return 4

    def _parse_client_init(self, packet) -> int:
        log("_parse_client_init(%s)", packet)
        ci_size = struct.calcsize(CLIENT_INIT)
        if len(packet) < ci_size:
            return 0
        # the last item in client init is the length of the session name:
        client_init = struct.unpack(CLIENT_INIT, packet[:ci_size])
        name_size = client_init[-1]
        # do we have enough to parse that too?
        if len(packet) < ci_size + name_size:
            return 0
        # w, h, bpp, depth, bigendian, truecolor, rmax, gmax, bmax, rshift, bshift, gshift = client_init[:12]
        w, h, bpp, depth, bigendian, truecolor = client_init[:6]
        sn = packet[ci_size:ci_size + name_size]
        try:
            session_name = sn.decode("utf8")
        except UnicodeDecodeError:
            session_name = sn.decode("latin1")
        log.info(f"RFB server session {session_name!r}: {w}x{h} {depth} bits")
        log(f"bpp={bpp}, bigendian={bool(bigendian)}")
        if not truecolor:
            self.invalid("server is not true color", packet)
            return 0
        # simulate hello:
        self._process_packet_cb(self, ["hello", {
            "session-name": session_name,
            "desktop_size": (w, h),
            "protocol": "rfb",
        }])
        # simulate an xpra window packet:
        metadata = {
            "title": session_name,
            "size-constraints": {
                "maximum-size": (w, h),
                "minimum-size": (w, h),
            },
            # "set-initial-position" : False,
            "window-type": ("NORMAL",),
            "has-alpha": False,
            # "decorations" : True,
            "content-type": "desktop",
        }
        client_properties = {}
        self._process_packet_cb(self, ["new-window", WID, 0, 0, w, h, metadata, client_properties])
        self._packet_parser = self._parse_rfb_packet
        self.send_set_encodings()

        # self.send_refresh_request(0, 0, 0, w, h)

        def request_refresh() -> bool:
            self.send_refresh_request(0, 0, 0, w, h)
            return True

        GLib.timeout_add(1000, request_refresh)
        return ci_size + name_size

    def send_set_encodings(self) -> None:
        self.send_struct("!BBHi", RFBClientMessage.SetEncodings, 0, 1, RFBEncoding.RAW)

    def send_refresh_request(self, incremental, x, y, w, h) -> None:
        self.send_struct("!BBHHHH", RFBClientMessage.FramebufferUpdateRequest, incremental, x, y, w, h)

    def _parse_rfb_packet(self, packet) -> int:
        log("parse_rfb_packet(%s)", repr_ellipsized(packet))
        if len(packet) <= 4:
            return 0
        if packet[:2] != struct.pack(b"!BB", 0, 0):
            self.invalid("unknown packet", packet)
            return 0
        self.rectangles = struct.unpack(b"!H", packet[2:4])[0]
        log("%i rectangles coming up", self.rectangles)
        if self.rectangles > 0:
            self._packet_parser = self._parse_rectangle
        return 4

    def _parse_rectangle(self, packet) -> int:
        header_size = struct.calcsize(b"!HHHHi")
        if len(packet) <= header_size:
            return 0
        x, y, w, h, encoding = struct.unpack(b"!HHHHi", packet[:header_size])
        if encoding != RFBEncoding.RAW:
            self.invalid(f"invalid encoding: {encoding}", packet)
            return 0
        if len(packet) < header_size + w * h * 4:
            return 0
        log("screen update: %s", (x, y, w, h))
        pixels = packet[header_size:header_size + w * h * 4]
        draw = ["draw", WID, x, y, w, h, "rgb32", pixels, 0, w * 4, {}]
        self._process_packet_cb(self, draw)
        self.rectangles -= 1
        if self.rectangles == 0:
            self._packet_parser = self._parse_rfb_packet
        return header_size + w * h * 4

    def send_struct(self, fmt, *args) -> None:
        packet = struct.pack(fmt, *args)
        self.send(packet)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.
import os
from typing import TypeAlias

from xpra.auth.sys_auth_base import log, DEFAULT_UID, DEFAULT_GID
from xpra.os_util import POSIX

SessionData: TypeAlias = tuple[int, int, list[str], dict[str, str], dict[str, str]]


def parse_uid(v) -> int:
    if v:
        try:
            return int(v)
        except (TypeError, ValueError):
            log(f"uid {v!r} is not an integer")
    if POSIX:
        try:
            import pwd  # pylint: disable=import-outside-toplevel
            return pwd.getpwnam(v or DEFAULT_UID).pw_uid
        except Exception as e:
            log(f"parse_uid({v})", exc_info=True)
            log.error(f"Error: cannot find uid of {v!r}: {e}")
        return os.getuid()
    return -1


def parse_gid(v) -> int:
    if v:
        try:
            return int(v)
        except (TypeError, ValueError):
            log(f"gid {v!r} is not an integer")
    if POSIX:
        try:
            import grp  # pylint: disable=import-outside-toplevel
            return grp.getgrnam(v or DEFAULT_GID).gr_gid
        except Exception as e:
            log(f"parse_gid({v})", exc_info=True)
            log.error(f"Error: cannot find gid of {v!r}: {e}")
        return os.getgid()
    return -1



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

# Augments the win32_NotifyIcon "system tray" support class
# with methods for integrating with win32_balloon and the popup menu

from time import monotonic
from ctypes import wintypes, byref

from xpra.platform.win32 import constants as win32con
from xpra.platform.win32.gui import EnumDisplayMonitors, GetMonitorInfo
from xpra.platform.win32.NotifyIcon import win32NotifyIcon
from xpra.platform.win32.events import get_win32_event_listener
from xpra.platform.win32.common import GetSystemMetrics, GetCursorPos
from xpra.client.gui.tray_base import TrayBase
from xpra.platform.paths import get_icon_filename
from xpra.log import Logger

log = Logger("tray", "win32")


class Win32Tray(TrayBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.calculate_offset()
        self.default_icon_extension = "ico"
        icon_filename = get_icon_filename(self.default_icon_filename, "ico")
        self.tray_widget = win32NotifyIcon(self.app_id, title=self.tooltip,
                                           move_callback=self.move_cb,
                                           click_callback=self.click_cb,
                                           exit_callback=self.exit_cb,
                                           iconPathName=icon_filename)
        el = get_win32_event_listener()
        if el:
            el.add_event_callback(win32con.WM_DISPLAYCHANGE, self.calculate_offset)

    def show(self):
        """ we can't hide or show the tray on MS Windows """

    def hide(self):
        """ we can't hide or show the tray on MS Windows """

    def get_size(self):
        w = GetSystemMetrics(win32con.SM_CXSMICON)
        h = GetSystemMetrics(win32con.SM_CYSMICON)
        return w, h

    def getHWND(self):
        if self.tray_widget is None:
            return None
        return self.tray_widget.hwnd

    def cleanup(self):
        log("Win32Tray.cleanup() tray_widget=%s", self.tray_widget)
        if self.tray_widget:
            self.tray_widget.close()
            self.tray_widget = None
        el = get_win32_event_listener()
        if el:
            el.remove_event_callback(win32con.WM_DISPLAYCHANGE, self.calculate_offset)
        log("Win32Tray.cleanup() ended")

    def calculate_offset(self, *_args):
        # GTK returns coordinates as unsigned ints, but win32 can give us negative coordinates!
        self.offset_x = 0
        self.offset_y = 0
        try:
            for m in EnumDisplayMonitors():
                mi = GetMonitorInfo(m)
                mx1, my1, _, _ = mi['Monitor']
                self.offset_x = max(self.offset_x, -mx1)
                self.offset_y = max(self.offset_y, -my1)
        except Exception as e:
            log("EnumDisplayMonitors()", exc_info=True)
            log.warn("Warning: failed to query monitors: %s", e)
        log("calculate_offset() x=%i, y=%i", self.offset_x, self.offset_y)

    def set_tooltip(self, tooltip=None):
        if self.tray_widget:
            self.tray_widget.set_tooltip(tooltip)

    def set_icon_from_data(self, pixels, has_alpha, w, h, rowstride, options=None):
        if self.tray_widget:
            self.tray_widget.set_icon_from_data(pixels, has_alpha, w, h, rowstride, options)
            self.icon_timestamp = monotonic()

    def do_set_icon_from_file(self, filename):
        if self.tray_widget:
            self.tray_widget.set_icon(filename)
            self.icon_timestamp = monotonic()

    def set_blinking(self, on):
        if self.tray_widget:
            self.tray_widget.set_blinking(on)

    def get_geometry(self):
        tw = self.tray_widget
        if tw:
            geom = tw.get_geometry()
            if geom:
                return geom
        return self.geometry_guess

    def move_cb(self, *args):
        pos = wintypes.POINT()
        GetCursorPos(byref(pos))
        x, y = pos.x, pos.y
        w = GetSystemMetrics(win32con.SM_CXSMICON)
        h = GetSystemMetrics(win32con.SM_CYSMICON)
        x += self.offset_x
        y += self.offset_y
        log("move_cb%s x=%s, y=%s, size=%i, %i", args, x, y, w, h)
        self.recalculate_geometry(x, y, w, h)
        if self.mouseover_cb:
            self.mouseover_cb(x, y)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2016 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest
from time import monotonic

from xpra.codecs.image import ImageWrapper
from xpra.util.env import envbool

SHOW_PERF = envbool("XPRA_SHOW_PERF")


class TestImageWrapper(unittest.TestCase):

    def test_sub_image(self):
        X = 0
        Y = 0
        W = 640
        H = 480
        D = 24
        buf = bytearray(W*H*4)
        #the pixel value is derived from the sum of its coordinates (modulo 256)
        for x in range(W):
            for y in range(H):
                #4 bytes per pixel:
                for i in range(4):
                    buf[y*(W*4) + x*4 + i] = (x+y) % 256
        img = ImageWrapper(X, Y, W, H, buf, "RGBX", D, W*4, planes=ImageWrapper.PACKED, thread_safe=True)
        #verify attributes:
        assert repr(img)
        assert img.get_x()==X
        assert img.get_y()==Y
        assert img.get_target_x()==X
        assert img.get_target_y()==Y
        assert img.get_width()==W
        assert img.get_height()==H
        assert img.get_depth()==D
        assert img.get_bytesperpixel()==4
        assert img.get_rowstride()==W*4
        assert img.get_size()==W*4*H
        assert img.has_pixels()
        assert len(img.get_geometry())==5
        assert img.get_pixel_format()=="RGBX"
        assert img.get_planes()==ImageWrapper.PACKED
        #print("image pixels head=%s" % (binascii.hexlify(img.get_pixels()[:128]), ))
        for x in range(3):
            SW, SH = 6, 6
            sub = img.get_sub_image(x, 0, SW, SH)
            #print("%s.get_sub_image%s=%s" % (img, (x, 0, SW, SH), sub))
            #this is only true for the pure python ImageWrapper:
            #(the X11 image wrapper references the same underlying XShm pixels, with the same rowstride)
            assert sub.get_rowstride()==(SW*4)
            sub_buf = sub.get_pixels()
            #print("pixels for %ix%i: %i" % (SW, SH, len(sub_buf)))
            #print("pixels=%s" % (binascii.hexlify(sub_buf), ))
            #verify that the pixels are set to 1 / 0:
            for y in range(SH):
                v = (x+y)%256
                for i in range(4):
                    av = sub_buf[y*(SW*4)+i]
                    if av!=v:
                        raise Exception("""expected value %#x for pixel (0, %i)
                                        of sub-image %s at (%i, 0), but got %#x""" % (v, y, sub, x, av))
        start = monotonic()
        copy = img.get_sub_image(0, 0, W, H)
        end = monotonic()
        if SHOW_PERF:
            print("image wrapper full %ix%i copy speed: %iMB/s" % (W, H, (W*4*H)/(end-start)/1024/1024))
        assert copy.get_pixels()==img.get_pixels()
        total = 0
        N = 10
        for i in range(N):
            region = (W//4-N//2+i, H//4-N//2+i, W//2, H//2)
            start = monotonic()
            copy = img.get_sub_image(*region)
            end = monotonic()
            total += end-start
        if SHOW_PERF:
            print("image wrapper sub image %ix%i copy speed: %iMB/s" % (W//2, H//2, N*(W//2*4*H//2)/total/1024/1024))
        #invalid sub-image should fail:
        for x, y, w, h in (
            (0, 0, 0, 1),
            (0, 0, 1, 0),
            (1, 0, W, 1),
            (0, 1, 1, H),
            ):
            try:
                img.get_sub_image(x, y, w, h)
            except Exception:
                pass
            else:
                raise Exception("sub image of %s with coords %s should have failed" % (img, (x, y, w, h)))
        #freeze is a no-op in the default wrapper:
        assert img.freeze() is False
        img.clone_pixel_data()

    def test_restride(self):
        #restride of planar is not supported:
        img = ImageWrapper(0, 0, 1, 1, ["0"*10, "0"*10, "0"*10, "0"*10],
                           "YUV420P", 24, 10, 3, planes=ImageWrapper.PLANAR_4)
        img.set_planes(ImageWrapper.PLANAR_3)
        img.clone_pixel_data()
        assert img.may_restride() is False
        img = ImageWrapper(0, 0, 1, 1, memoryview(b"0"*4), "BGRA", 24, 4, 4, planes=ImageWrapper.PACKED)
        img.clone_pixel_data()
        assert img.may_restride() is False
        img = ImageWrapper(0, 0, 10, 10, b"0"*40*10, "BGRA", 24, 40, 4, planes=ImageWrapper.PACKED)
        #restride bigger:
        img.restride(80)
        #change more attributes:
        img.set_timestamp(img.get_timestamp()+1)
        img.set_pixel_format("RGBA")
        img.set_palette(())
        img.set_pixels("1"*10)
        assert img.allocate_buffer(0, 1)==0
        assert img.get_palette()==()
        assert img.is_thread_safe()
        assert img.get_gpu_buffer() is None
        img.set_rowstride(20)


def main():
    unittest.main()

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest

from xpra.util.objects import AdHocStruct
from unit.server.mixins.servermixintest_util import ServerMixinTest


class WebcamMixinTest(ServerMixinTest):

    def test_windowserver(self):
        from xpra.server.mixins.window import WindowServer
        opts = AdHocStruct()
        opts.min_size = "10x10"
        opts.max_size = "16384x8192"
        def load_existing_windows():
            pass
        def _WindowServer():
            ws = WindowServer()
            ws.load_existing_windows = load_existing_windows
            return ws
        self._test_mixin_class(_WindowServer, opts)

def main():
    unittest.main()


if __name__ == '__main__':
    main()



# This file is part of Xpra.
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from typing import Any
from collections.abc import Sequence, Callable

from xpra.os_util import gi_import
from xpra.server.window import batch_config
from xpra.server.shadow.root_window_model import RootWindowModel
from xpra.scripts.config import InitExit
from xpra.platform.gui import get_native_notifier_classes, get_wm_name
from xpra.platform.paths import get_icon_dir
from xpra.server import features
from xpra.exit_codes import ExitCode
from xpra.net.common import PacketType
from xpra.exit_codes import ExitValue
from xpra.util.system import is_Wayland
from xpra.util.env import envint, envbool
from xpra.util.str_fn import csv
from xpra.common import NotificationID, ConnectionMessage
from xpra.log import Logger

GLib = gi_import("GLib")

log = Logger("shadow")
notifylog = Logger("notify")
mouselog = Logger("mouse")
cursorlog = Logger("cursor")

NATIVE_NOTIFIER = envbool("XPRA_NATIVE_NOTIFIER", True)
POLL_POINTER = envint("XPRA_POLL_POINTER", 20)
CURSORS = envbool("XPRA_CURSORS", True)
SAVE_CURSORS = envbool("XPRA_SAVE_CURSORS", False)
NOTIFY_STARTUP = envbool("XPRA_SHADOW_NOTIFY_STARTUP", True)

SHADOWSERVER_BASE_CLASS: type = object
if features.rfb:
    from xpra.server.rfb.server import RFBServer
    SHADOWSERVER_BASE_CLASS = RFBServer


def try_setup_capture(backends: dict[str, Callable], backend: str, *args):
    backend = backend.lower()
    if backend != "auto":
        if backend not in backends:
            raise InitExit(ExitCode.UNSUPPORTED,
                           f"invalid capture backend {backend!r}, should be one of: %s" % csv(backends.keys()))
        backends = {backend: backends[backend]}
    log(f"setup_capture() will try {backends.keys()}")
    for backend, setup_fn in backends.items():
        try:
            log(f"{backend!r}: {setup_fn}{args}")
            capture = setup_fn(*args)
            if not capture:
                log(f"backend {backend!r} is unable to capture")
                continue
            return capture
        except ImportError as e:
            log(f"backend {backend!r} is not installed: {e}")
        except Exception as e:
            log(f"{setup_fn}{args}", exc_info=True)
            log.warn(f"Warning: {backend!r} failed to setup screen capture:")
            log.warn(" %s", e)
    raise InitExit(ExitCode.UNSUPPORTED, f"failed to setup screen capture backend {backend!r}")


class ShadowServerBase(SHADOWSERVER_BASE_CLASS):
    # 20 fps unless the client specifies more:
    DEFAULT_REFRESH_RATE: int = 20

    def __init__(self, root_window, capture=None):
        # noinspection PyArgumentList
        super().__init__()
        self.capture = capture
        self.root = root_window
        self.window_matches: list[str] = []
        self.mapped = []
        self.pulseaudio: bool = False
        self.sharing: bool | None = None
        self.refresh_delay: int = 1000 // self.DEFAULT_REFRESH_RATE
        self.refresh_timer: int = 0
        self.notifications: bool = False
        self.notifier = None
        self.pointer_last_position = None
        self.pointer_poll_timer = 0
        self.last_cursor_data = None
        self.session_name = "shadow"
        self.keyboard_config = None
        batch_config.ALWAYS = True  # always batch

    def init(self, opts) -> None:
        if SHADOWSERVER_BASE_CLASS is not object:
            # RFBServer:
            SHADOWSERVER_BASE_CLASS.init(self, opts)
        self.notifications = bool(opts.notifications)
        if self.notifications:
            self.make_notifier()
        log("init(..) session_name=%s", opts.session_name)
        if opts.session_name:
            self.session_name = opts.session_name
        else:
            self.guess_session_name()

    def run(self) -> ExitValue:
        if NOTIFY_STARTUP:
            GLib.timeout_add(1000, self.notify_startup_complete)
        return super().run()

    def cleanup(self) -> None:
        for wid in self.mapped:
            self.stop_refresh(wid)
        self.cleanup_notifier()
        self.cleanup_capture()

    def cleanup_capture(self) -> None:
        capture = self.capture
        if capture:
            self.capture = None
            capture.clean()

    def guess_session_name(self, procs=None) -> None:
        log("guess_session_name(%s)", procs)
        self.session_name = get_wm_name()  # pylint: disable=assignment-from-none
        log("get_wm_name()=%s", self.session_name)

    def get_server_mode(self) -> str:
        return "shadow"

    def print_screen_info(self) -> None:
        if not features.display or not self.root:
            return
        w, h = self.root.get_geometry()[2:4]
        dinfo = ""
        if is_Wayland():
            wdisplay = os.environ.get("WAYLAND_DISPLAY", "").replace("wayland-", "")
            if wdisplay:
                dinfo = f"Wayland display {wdisplay}"
        else:
            display = os.environ.get("DISPLAY")
            if display:
                dinfo = f"X11 display {display}"
        self.do_print_screen_info(dinfo, w, h)

    def do_print_screen_info(self, display: str, w: int, h: int) -> None:
        if display or w or h:
            if display:
                msg = f" on display {display!r}"
                if w and h:
                    msg += f" of size {w}x{h}"
            else:
                msg = f" on display of size {w}x{h}"
            log.info(msg)
        if self.window_matches:
            return
        try:
            count = len(self._id_to_window)
        except AttributeError as e:
            log(f"no screen info: {e}")
            return
        if count > 1:
            log.info(f" with {count} monitors:")
            for window in self._id_to_window.values():
                title = window.get_property("title")
                x, y, w, h = window.geometry
                log.info("  %-16s %4ix%-4i at %4i,%-4i", title, w, h, x, y)

    def apply_refresh_rate(self, ss) -> None:
        rrate = super().apply_refresh_rate(ss)
        if rrate > 0:
            # adjust refresh delay to try to match:
            self.set_refresh_delay(max(10, 1000 // rrate))

    def make_hello(self, _source) -> dict[str, Any]:
        return {"shadow": True}

    def get_info(self, _proto=None, *args) -> dict[str, Any]:
        info = {
            "sharing": self.sharing is not False,
            "refresh-delay": self.refresh_delay,
        }
        if self.pointer_last_position:
            info["pointer-last-position"] = self.pointer_last_position
        return info

    def get_window_position(self, _window) -> tuple[int, int]:
        # we export the whole desktop as a window:
        return 0, 0

    def _keys_changed(self) -> None:
        from xpra.server.mixins.input import InputServer
        if isinstance(self, InputServer):
            InputServer._keys_changed(self)
            from xpra.platform.keyboard import Keyboard
            log.info("the keymap has been changed: %s", Keyboard().get_layout_spec()[0])

    ############################################################################
    # notifications
    def cleanup_notifier(self) -> None:
        n = self.notifier
        if n:
            self.notifier = None
            n.cleanup()

    def notify_setup_error(self, exception) -> None:
        notifylog("notify_setup_error(%s)", exception)
        notifylog.info("notification forwarding is not available")
        if str(exception).endswith("is already claimed on the session bus"):
            log.info(" the interface is already claimed")

    def make_notifier(self) -> None:
        nc = self.get_notifier_classes()
        notifylog("make_notifier() notifier classes: %s", nc)
        for nclass in nc:
            try:
                self.notifier = nclass()
                notifylog("notifier=%s", self.notifier)
                break
            except Exception:
                notifylog("failed to instantiate %s", nclass, exc_info=True)

    def get_notifier_classes(self) -> list[type]:
        # subclasses will generally add their toolkit specific variants
        # by overriding this method
        # use the native ones first:
        if not NATIVE_NOTIFIER:
            return []
        return get_native_notifier_classes()

    def notify_new_user(self, ss) -> None:
        # overridden here so that we can show the notification
        # directly on the screen we shadow
        notifylog("notify_new_user(%s) notifier=%s", ss, self.notifier)
        if self.notifier:
            tray = self.get_notification_tray()  # pylint: disable=assignment-from-none
            nid = NotificationID.NEW_USER
            title = "User '%s' connected to the session" % (ss.name or ss.username or ss.uuid)
            body = "\n".join(ss.get_connect_info())
            actions: list = []
            hints: dict[str, Any] = {}
            icon_filename = os.path.join(get_icon_dir(), "user.png")
            from xpra.notifications.common import parse_image_path
            icon = parse_image_path(icon_filename)
            self.notifier.show_notify("", tray, nid, "Xpra", 0, "", title, body, actions, hints, 10 * 1000, icon)

    def get_notification_tray(self):
        return None

    def notify_startup_complete(self) -> None:
        self.do_notify_startup("Xpra shadow server is ready", replaces_nid=NotificationID.STARTUP)

    def do_notify_startup(self, title: str, body: str = "", replaces_nid: int | NotificationID = 0) -> None:
        # this is overridden here so that we can show the notification
        # directly on the screen we shadow
        notifylog("do_notify_startup%s", (title, body, replaces_nid))
        if self.notifier:
            tray = self.get_notification_tray()  # pylint: disable=assignment-from-none
            actions = []
            hints = {}
            icon_filename = os.path.join(get_icon_dir(), "server-connected.png")
            from xpra.notifications.common import parse_image_path
            icon = parse_image_path(icon_filename)
            self.notifier.show_notify("", tray, NotificationID.STARTUP, "Xpra", replaces_nid, "",
                                      title, body, actions, hints, 10 * 1000, icon)

    ############################################################################
    # refresh

    def start_refresh(self, wid: int) -> None:
        log("start_refresh(%i) mapped=%s, timer=%s", wid, self.mapped, self.refresh_timer)
        if wid not in self.mapped:
            self.mapped.append(wid)
        self.start_refresh_timer()
        self.start_poll_pointer()

    def start_refresh_timer(self) -> None:
        if not self.refresh_timer:
            self.refresh_timer = GLib.timeout_add(self.refresh_delay, self.refresh)

    def set_refresh_delay(self, v: int) -> None:
        assert 0 < v < 10000
        self.refresh_delay = v
        if self.mapped:
            self.cancel_refresh_timer()
            for wid in self.mapped:
                self.start_refresh(wid)

    def stop_refresh(self, wid: int) -> None:
        log("stop_refresh(%i) mapped=%s", wid, self.mapped)
        try:
            self.mapped.remove(wid)
        except ValueError:
            pass
        if not self.mapped:
            self.no_windows()

    def no_windows(self) -> None:
        self.cancel_refresh_timer()
        self.cancel_poll_pointer()

    def cancel_refresh_timer(self) -> None:
        t = self.refresh_timer
        log("cancel_refresh_timer() timer=%s", t)
        if t:
            self.refresh_timer = 0
            GLib.source_remove(t)

    def refresh(self) -> bool:
        raise NotImplementedError()

    ############################################################################
    # pointer polling

    def get_pointer_position(self):
        raise NotImplementedError()

    def start_poll_pointer(self) -> None:
        log("start_poll_pointer() pointer_poll_timer=%s, input_devices=%s, POLL_POINTER=%s",
            self.pointer_poll_timer, features.input_devices, POLL_POINTER)
        if self.pointer_poll_timer:
            self.cancel_poll_pointer()
        if features.input_devices and POLL_POINTER > 0:
            self.pointer_poll_timer = GLib.timeout_add(POLL_POINTER, self.poll_pointer)

    def cancel_poll_pointer(self) -> None:
        ppt = self.pointer_poll_timer
        log("cancel_poll_pointer() pointer_poll_timer=%s", ppt)
        if ppt:
            self.pointer_poll_timer = 0
            GLib.source_remove(ppt)

    def poll_pointer(self) -> bool:
        self.poll_pointer_position()
        if CURSORS:
            self.poll_cursor()
        return True

    def poll_pointer_position(self) -> None:
        x, y = self.get_pointer_position()
        if self.pointer_last_position == (x, y):
            mouselog("poll_pointer_position() unchanged position=%s", (x, y))
            return
        self.pointer_last_position = (x, y)
        rwm = None
        wid = None
        rx, ry = 0, 0
        # find the window model containing the pointer:
        for wid, window in self._id_to_window.items():
            wx, wy, ww, wh = window.geometry
            if wx <= x < (wx + ww) and wy <= y < (wy + wh):
                rwm = window
                rx = x - wx
                ry = y - wy
                break
        if not rwm:
            mouselog("poll_pointer_position() model not found for position=%s", (x, y))
            return
        mouselog("poll_pointer_position() wid=%i, position=%s, relative=%s", wid, (x, y), (rx, ry))
        for ss in self._server_sources.values():
            um = getattr(ss, "update_mouse", None)
            if um:
                um(wid, x, y, rx, ry)

    def poll_cursor(self) -> None:
        prev = self.last_cursor_data
        curr = self.do_get_cursor_data()  # pylint: disable=assignment-from-none
        self.last_cursor_data = curr

        def cmpv(lcd: Sequence | None) -> tuple[Any, ...]:
            if not lcd:
                return ()
            v = lcd[0]
            if v and len(v) > 2:
                return tuple(v[2:])
            return ()

        if cmpv(prev) != cmpv(curr):
            fields = ("x", "y", "width", "height", "xhot", "yhot", "serial", "pixels", "name")
            if len(prev or []) == len(curr or []) and len(prev or []) == len(fields):
                diff = []
                for i, prev_value in enumerate(prev):
                    if prev_value != curr[i]:
                        diff.append(fields[i])
                cursorlog("poll_cursor() attributes changed: %s", diff)
            if SAVE_CURSORS and curr:
                ci = curr[0]
                if ci:
                    w = ci[2]
                    h = ci[3]
                    serial = ci[6]
                    pixels = ci[7]
                    cursorlog("saving cursor %#x with size %ix%i, %i bytes", serial, w, h, len(pixels))
                    from PIL import Image
                    img = Image.frombuffer("RGBA", (w, h), pixels, "raw", "BGRA", 0, 1)
                    img.save("cursor-%#x.png" % serial, format="PNG")
            for ss in self.window_sources():
                ss.send_cursor()

    def do_get_cursor_data(self):
        # this method is overridden in subclasses with platform specific code
        return None

    def get_cursor_data(self):
        # return cached value we get from polling:
        return self.last_cursor_data

    ############################################################################

    def sanity_checks(self, _proto, c) -> bool:
        server_uuid = c.strget("server_uuid")
        if server_uuid:
            if server_uuid == self.uuid:
                log.warn("Warning: shadowing your own display can be quite confusing")
                clipboard = self._clipboard_helper and c.boolget("clipboard", True)
                if clipboard:
                    log.warn("Warning: clipboard sharing cannot be enabled!")
                    log.warn(" consider using the --no-clipboard option")
                    c["clipboard"] = False
            else:
                log.warn("This client is running within the Xpra server %s", server_uuid)
        return True

    def parse_screen_info(self, ss):
        try:
            log.info(" client root window size is %sx%s", *ss.desktop_size)
        except Exception:
            log.info(" unknown client desktop size")
        self.apply_refresh_rate(ss)
        return self.get_root_window_size()

    def _process_desktop_size(self, proto, packet: PacketType) -> None:
        # just record the screen size info in the source
        ss = self.get_server_source(proto)
        if ss and len(packet) >= 4:
            ss.set_screen_sizes(packet[3])

    def set_keyboard_repeat(self, key_repeat) -> None:
        """ don't override the existing desktop """
        pass  # pylint: disable=unnecessary-pass

    def set_keymap(self, server_source, force=False) -> None:
        log("set_keymap%s", (server_source, force))
        log.info("shadow server: setting default keymap translation")
        self.keyboard_config = server_source.set_default_keymap()

    def load_existing_windows(self) -> None:
        self.mmap_min_size = 1024 * 1024 * 4 * 2
        for i, model in enumerate(self.makeRootWindowModels()):
            log(f"load_existing_windows() root window model {i} : {model}")
            self._add_new_window(model)
            # at least big enough for 2 frames of BGRX pixel data:
            w, h = model.get_dimensions()
            self.mmap_min_size = max(self.mmap_min_size, w * h * 4 * 2)

    def makeRootWindowModels(self) -> list:
        return [RootWindowModel(self.root)]

    def send_initial_windows(self, ss, sharing: bool = False) -> None:
        log("send_initial_windows(%s, %s) will send: %s", ss, sharing, self._id_to_window)
        for wid in sorted(self._id_to_window.keys()):
            window = self._id_to_window[wid]
            w, h = window.get_dimensions()
            client_props = self.client_properties.get(wid, {}).get(ss.uuid, {})
            ss.new_window("new-window", wid, window, 0, 0, w, h, client_props)

    def _add_new_window(self, window) -> None:
        self._add_new_window_common(window)
        if window.get("override-redirect", False):
            self._send_new_or_window_packet(window)
        else:
            self._send_new_window_packet(window)

    def _send_new_window_packet(self, window) -> None:
        geometry = window.get_geometry()
        self._do_send_new_window_packet("new-window", window, geometry)

    def _send_new_or_window_packet(self, window) -> None:
        geometry = window.get_property("geometry")
        self._do_send_new_window_packet("new-override-redirect", window, geometry)

    def process_window_common(self, wid: int):
        return self._id_to_window.get(wid)

    def _process_map_window(self, proto, packet: PacketType) -> None:
        wid, x, y, width, height = packet[1:6]
        window = self.process_window_common(wid)
        if not window:
            # already gone
            return
        self._window_mapped_at(proto, wid, window, (x, y, width, height))
        self.refresh_window_area(window, 0, 0, width, height)
        if len(packet) >= 7:
            self._set_client_properties(proto, wid, window, packet[6])
        self.start_refresh(wid)

    def _process_unmap_window(self, proto, packet: PacketType) -> None:
        wid = packet[1]
        window = self.process_window_common(wid)
        if not window:
            # already gone
            return
        self._window_mapped_at(proto, wid, window)
        # TODO: deal with more than one window / more than one client
        # and stop refresh if all the windows are unmapped everywhere
        if len(self._server_sources) <= 1 and len(self._id_to_window) <= 1:
            self.stop_refresh(wid)

    def _process_configure_window(self, proto, packet: PacketType) -> None:
        wid, x, y, w, h = packet[1:6]
        window = self.process_window_common(wid)
        if not window:
            # already gone
            return
        self._window_mapped_at(proto, wid, window, (x, y, w, h))
        self.refresh_window_area(window, 0, 0, w, h)
        if len(packet) >= 7:
            self._set_client_properties(proto, wid, window, packet[6])

    def _process_close_window(self, proto, packet: PacketType) -> None:
        wid = packet[1]
        window = self.process_window_common(wid)
        if not window:
            # already gone
            return
        # FIXME: with multiple windows / clients,
        # we have to keep track of mappings!
        if len(self._window_to_id) == 1:
            self.disconnect_client(proto, ConnectionMessage.DONE, "closed the only window")

    def do_make_screenshot_packet(self):
        raise NotImplementedError()

    def make_dbus_server(self):
        from xpra.server.dbus.shadow_server import Shadow_DBUS_Server
        return Shadow_DBUS_Server(self, os.environ.get("DISPLAY", "").lstrip(":"))



# This file is part of Xpra.
# Copyright (C) 2010 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

# A tray implemented using Gtk.StatusIcon

import os
from time import time, monotonic

from xpra.os_util import WIN32, OSX, POSIX, gi_import
from xpra.util.env import envbool, ignorewarnings
from xpra.client.gui.tray_base import TrayBase, log
from xpra.gtk.util import get_default_root_window
from xpra.gtk.pixbuf import get_icon_from_file, get_pixbuf_from_data

Gdk = gi_import("Gdk")
Gtk = gi_import("Gtk")
GdkPixbuf = gi_import("GdkPixbuf")

ORIENTATION = {
    Gtk.Orientation.HORIZONTAL: "HORIZONTAL",
    Gtk.Orientation.VERTICAL: "VERTICAL",
}

GUESS_GEOMETRY = WIN32 or OSX
GUESS_GEOMETRY = envbool("XPRA_GUESS_ICON_GEOMETRY", GUESS_GEOMETRY)
log("tray GUESS_GEOMETRY=%s", GUESS_GEOMETRY)
SAVE = envbool("XPRA_SAVE_SYSTRAY", False)


class GTKStatusIconTray(TrayBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tray_widget = Gtk.StatusIcon()
        ignorewarnings(self.tray_widget.set_tooltip_text, self.tooltip or "Xpra")
        self.tray_widget.connect('activate', self.activate_menu)
        self.tray_widget.connect('popup-menu', self.popup_menu)
        if self.size_changed_cb:
            self.tray_widget.connect('size-changed', self.size_changed_cb)
        if self.default_icon_filename:
            self.set_icon()
        ignorewarnings(self.tray_widget.set_visible, True)

    def may_guess(self) -> None:
        log("may_guess() GUESS_GEOMETRY=%s, current guess=%s", GUESS_GEOMETRY, self.geometry_guess)
        if GUESS_GEOMETRY:
            x, y = get_default_root_window().get_pointer()[-3:-1]
            w, h = self.get_size()
            self.recalculate_geometry(x, y, w, h)

    def activate_menu(self, widget) -> None:
        modifiers_mask = get_default_root_window().get_pointer()[-1]
        log("activate_menu(%s) modifiers_mask=%s", widget, modifiers_mask)
        if (modifiers_mask & Gdk.ModifierType.SHIFT_MASK) ^ OSX:
            self.handle_click(3)
        else:
            self.handle_click(1)

    def popup_menu(self, widget, button: int, event_time, *args) -> None:
        modifiers_mask = get_default_root_window().get_pointer()[-1]
        log("popup_menu(%s, %s, %s, %s) modifiers_mask=%s", widget, button, event_time, args, modifiers_mask)
        if (modifiers_mask & Gdk.ModifierType.SHIFT_MASK) ^ OSX:
            self.handle_click(1)
        else:
            self.handle_click(3)

    def handle_click(self, button: int, event_time=0) -> None:
        log("handle_click(%i, %i)", button, event_time)
        self.may_guess()
        if self.click_cb:
            self.click_cb(button, 1, event_time)
            self.click_cb(button, 0, event_time)

    def hide(self) -> None:
        log("%s.set_visible(False)", self.tray_widget)
        if self.tray_widget:
            self.tray_widget.set_visible(False)

    def show(self) -> None:
        log("%s.set_visible(True)", self.tray_widget)
        if self.tray_widget:
            self.tray_widget.set_visible(True)

    def get_orientation(self) -> str:
        if not self.tray_widget:
            return ""
        ag = self.tray_widget.get_geometry()
        if ag is None:
            return ""
        gtk_orientation = ag[-1]
        return ORIENTATION.get(gtk_orientation, "")

    def get_geometry(self):
        assert self.tray_widget
        # on X11, if we don't have an `xid`, don't bother querying its geometry,
        # as this would trigger some ugly GTK warnings we can do nothing about
        if POSIX and os.environ.get("DISPLAY") and ignorewarnings(self.tray_widget.get_x11_window_id) == 0:
            ag = None
        else:
            ag = ignorewarnings(self.tray_widget.get_geometry)
            log("GTKStatusIconTray.get_geometry() %s.get_geometry()=%s", self.tray_widget, ag)
        if ag is None:
            if not self.geometry_guess:
                self.may_guess()
            # probably win32 or OSX, gnome-shell or KDE5..
            log("GTKStatusIconTray.get_geometry() no geometry value available, returning guess: %s",
                self.geometry_guess)
            return self.geometry_guess or (0, 0, 0, 0)
        # `GTK3` adds an extra argument.. at the beginning,
        # so we index from the end of the array:
        geom = ag[-2]
        x, y, w, h = geom.x, geom.y, geom.width, geom.height
        log("GTKStatusIconTray.get_geometry() geometry area rectangle=%s", (x, y, w, h))
        if x == 0 and y == 0 and w == 0 and h == 0 and self.geometry_guess:
            return self.geometry_guess
        if x == 0 and y == 0 and w == 200 and h == 200:
            # this isn't right, take a better guess, at least for the size:
            w = 48
            h = 48
        return x, y, w, h

    def get_size(self):
        s = max(8, min(256, self.tray_widget.get_size()))
        return [s, s]

    def set_tooltip(self, tooltip: str = "") -> None:
        if self.tray_widget:
            ignorewarnings(self.tray_widget.set_tooltip_text, tooltip or "Xpra")

    def set_blinking(self, on: bool) -> None:
        # no longer supported with GTK3
        pass

    def set_icon_from_data(self, pixels, has_alpha: bool, w: int, h: int, rowstride: int, _options=None) -> None:
        tray_icon = get_pixbuf_from_data(pixels, has_alpha, w, h, rowstride)
        self.set_icon_from_pixbuf(tray_icon)

    def do_set_icon_from_file(self, filename: str) -> None:
        tray_icon = get_icon_from_file(filename)
        self.set_icon_from_pixbuf(tray_icon)

    def set_icon_from_pixbuf(self, tray_icon) -> None:
        if not tray_icon or not self.tray_widget:
            return
        tw, th = self.get_geometry()[2:4]
        if (tw <= 2 or th <= 2) or (tw == 200 and th == 200):
            log("bogus tray icon size: %ix%i", tw, th)
            tw = th = 48
        w = tray_icon.get_width()
        h = tray_icon.get_height()
        log("set_icon_from_pixbuf(%s) geometry=%s, icon size=%s", tray_icon, self.get_geometry(), (w, h))
        if tw != w or th != h:
            tray_icon = tray_icon.scale_simple(tw, th, GdkPixbuf.InterpType.HYPER)
            log("tray icon scaled to %ix%i", tw, th)
        if SAVE:
            filename = "./statusicon-%s.png" % time()
            tray_icon.savev(filename, "png")
            log.info("statusicon tray saved to %s", filename)
        ignorewarnings(self.tray_widget.set_from_pixbuf, tray_icon)
        self.icon_timestamp = monotonic()


def main() -> None:
    log.enable_debug()
    GLib = gi_import("GLib")
    log.enable_debug()
    s = GTKStatusIconTray(None, None, None, "test", "xpra.png", None, None, None, Gtk.main_quit)
    GLib.timeout_add(1000 * 2, s.set_blinking, True)
    GLib.timeout_add(1000 * 5, s.set_blinking, False)
    GLib.timeout_add(1000 * 30, Gtk.main_quit)
    Gtk.main()


if __name__ == "__main__":
    main()
